"""Defines the DrugBank ETL methods."""

import csv
from typing import Any

from tqdm import tqdm

from therapy.etl.base import Base
from therapy.schemas import NamespacePrefix, SourceMeta, SourceName


class DrugBank(Base):
    """Class for DrugBank ETL methods."""

    def _load_meta(self) -> None:
        """Add DrugBank metadata."""
        metadata = SourceMeta(
            data_license="CC0 1.0",
            data_license_url="https://creativecommons.org/publicdomain/zero/1.0/",
            version=self._version,
            data_url="https://go.drugbank.com/releases/latest#open-data",
            rdp_url="http://reusabledata.org/drugbank.html",
            data_license_attributes={
                "non_commercial": False,
                "share_alike": False,
                "attribution": False,
            },
        )
        self.database.add_source_metadata(SourceName.DRUGBANK, metadata)

    def _transform_data(self) -> None:
        """Transform the DrugBank source."""
        with self._data_file.open() as file:  # type: ignore
            reader = list(csv.reader(file))
            for row in tqdm(reader[1:], ncols=80, disable=self._silent):
                # get concept ID
                params: dict[str, Any] = {
                    "concept_id": f"{NamespacePrefix.DRUGBANK.value}:{row[0]}",
                }

                # get label
                label = row[2]
                if label:
                    params["label"] = label

                # get aliases
                aliases = [a for a in row[1].split(" | ") + row[5].split(" | ") if a]
                if aliases:
                    params["aliases"] = aliases

                # get CAS reference
                cas_ref = row[3]
                if cas_ref:
                    params["xrefs"] = [f"{NamespacePrefix.CHEMIDPLUS.value}:{row[3]}"]

                params["associated_with"] = []
                # get inchi key
                if len(row) >= 7:
                    inchi_key = row[6]
                    if inchi_key:
                        inchi_id = f"{NamespacePrefix.INCHIKEY.value}:{inchi_key}"
                        params["associated_with"].append(inchi_id)
                # get UNII id
                unii = row[4]
                if unii:
                    unii_id = f"{NamespacePrefix.UNII.value}:{unii}"
                    params["associated_with"].append(unii_id)

                self._load_therapy(params)



"""Test that the therapy normalizer works as intended for the RxNorm source."""

import isodate
import pytest

from therapy.etl import RxNorm
from therapy.schemas import MatchType, Therapy


@pytest.fixture(scope="module")
def rxnorm(test_source):
    """Provide test chembl query endpoint"""
    return test_source(RxNorm)


@pytest.fixture(scope="module")
def bifidobacterium_infantis():
    """Create bifidobacterium infantis drug fixture."""
    params = {
        "label": "Bifidobacterium infantis",
        "concept_id": "rxcui:100213",
        "aliases": [
            "Bifidobacterium infantis whole",
            "Bifidobacterium lactentis whole",
            "Bifidobacterium liberorum whole",
            "Bifidobacterium longum subsp. infantis",
            "Bifidobacterium longum subsp. infantis whole",
            "Bifidobacterium longum infantis whole",
            "bifidobacterium infantis",
            "Bifidobacterium longum bv. infantis whole",
        ],
        "approval_ratings": ["rxnorm_prescribable"],
        "xrefs": ["drugbank:DB14222"],
        "associated_with": ["mmsl:d07347", "unii:EF3U3H7I76", "mmsl:32421"],
        "trade_names": ["Align", "Evivo"],
    }
    return Therapy(**params)


@pytest.fixture(scope="module")
def cisplatin():
    """Create cisplatin drug fixture."""
    params = {
        "label": "cisplatin",
        "concept_id": "rxcui:2555",
        "aliases": [
            "(SP-4-2)-DIAMMINEDICHLOROPLATINUM",
            "CDDP",
            "CIS-DIAMMINEDICHLOROPLATINUM",
            "CIS-DIAMMINEDICHLOROPLATINUM II",
            "CISplatin",
            "Cis-DDP",
            "DDP",
            "PLATINUM, DIAMMINEDICHLORO-, (SP-4-2)-",
            "cis-diamminedichloroplatinum(II)",
        ],
        "approval_ratings": ["rxnorm_prescribable"],
        "xrefs": ["drugbank:DB00515", "drugbank:DB12117"],
        "associated_with": [
            "usp:m17910",
            "vandf:4018139",
            "unii:Q20Q21Q62J",
            "mmsl:d00195",
            "atc:L01XA01",
            "mmsl:31747",
            "mmsl:4456",
        ],
        "trade_names": ["Cisplatin", "Platinol"],
    }
    return Therapy(**params)


@pytest.fixture(scope="module")
def amiloride_hydrochloride():
    """Create amiloride hydrochloride drug fixture."""
    params = {
        "label": "amiloride hydrochloride",
        "concept_id": "rxcui:142424",
        "aliases": [
            "aMILoride hydrochloride",
        ],
        "approval_ratings": ["rxnorm_prescribable"],
        "xrefs": [],
        "associated_with": [
            "usp:m2650",
            "unii:FZJ37245UC",
            "mmsl:2658",
            "mmsl:4166",
            "vandf:4017935",
        ],
        "trade_names": [
            "AMILoride Hydrochloride",
            "Midamor",
            "Aridil",
            "Frusemek",
            "Moduret",
            "Moduretic",
            "Zida-Co",
        ],
    }
    return Therapy(**params)


@pytest.fixture(scope="module")
def amiloride():
    """Create amiloride drug fixture."""
    params = {
        "label": "amiloride",
        "concept_id": "rxcui:644",
        "aliases": [
            "3,5-diamino-N-carbamimidoyl-6-chloropyrazine-2-carboxamide",
            "aMILoride",
            "Amipramidin",
            "Amipramidine",
            "N-amidino-3,5-diamino-6-chloropyrazinecarboxamide",
            "Amyloride",
        ],
        "approval_ratings": ["rxnorm_prescribable"],
        "xrefs": ["drugbank:DB00594"],
        "associated_with": [
            "vandf:4019603",
            "unii:7DZO8EB0Z3",
            "mmsl:d00169",
            "atc:C03DB01",
        ],
        "trade_names": [
            "Amilamont",
        ],
    }
    return Therapy(**params)


@pytest.fixture(scope="module")
def timolol():
    """Create timolol drug fixture."""
    params = {
        "label": "timolol",
        "concept_id": "rxcui:10600",
        "aliases": [
            "(S)-1-(tert-butylamino)-3-[(4-morpholin-4-yl-1,2,5-thiadiazol-3-yl)oxy]propan-2-ol",
        ],
        "approval_ratings": ["rxnorm_prescribable"],
        "xrefs": ["drugbank:DB00373"],
        "associated_with": [
            "vandf:4019949",
            "unii:817W3C6175",
            "mmsl:d00139",
            "atc:C07AA06",
            "atc:S01ED01",
        ],
        "trade_names": [
            "Betimol",
            "Timoptic",
            "Istalol",
            "Combigan",
            "Cosopt",
            "Timoptol LA",
            "Blocadren",
            "Timolide",
            "Betim",
            "Glau-Opt",
            "Glaucol",
            "Timoptol",
        ],
    }
    return Therapy(**params)


@pytest.fixture(scope="module")
def lymphocyte():
    """Create lymphocyte immune globulin, anti-thymocyte globulin drug
    fixture.
    """
    params = {
        "label": "lymphocyte immune globulin, anti-thymocyte globulin",
        "concept_id": "rxcui:1011",
        "aliases": [
            "lymphocyte immune globulin, anti-thy (obs)",
        ],
        "approval_ratings": ["rxnorm_prescribable"],
        "xrefs": [],
        "associated_with": [
            "vandf:4022194",
            "vandf:4018097",
            "mmsl:d01141",
            "mmsl:5001",
        ],
        "trade_names": ["ATGAM", "Thymoglobulin"],
    }
    return Therapy(**params)


@pytest.fixture(scope="module")
def aspirin():
    """Create aspirin drug fixture."""
    params = {
        "label": "aspirin",
        "concept_id": "rxcui:1191",
        "aliases": [
            "Salicylic acid acetate",
            "Aspirin",
            "2-Acetoxybenzenecarboxylic acid",
            "ASA",
            "O-acetylsalicylic acid",
            "2-Acetoxybenzoic acid",
            "o-acetoxybenzoic acid",
            "o-carboxyphenyl acetate",
        ],
        "approval_ratings": ["rxnorm_prescribable"],
        "xrefs": ["drugbank:DB00945"],
        "associated_with": [
            "usp:m6240",
            "vandf:4017536",
            "mmsl:34512",
            "unii:R16CO5Y76E",
            "mmsl:244",
            "mmsl:4223",
            "mmsl:d00170",
            "atc:A01AD05",
            "atc:B01AC06",
            "atc:N02BA01",
        ],
        "trade_names": [],
    }
    return Therapy(**params)


@pytest.fixture(scope="module")
def mesna():
    """Create mesna drug fixture."""
    params = {
        "label": "mesna",
        "concept_id": "rxcui:44",
        "aliases": [],
        "approval_ratings": ["rxnorm_prescribable"],
        "xrefs": [],
        "associated_with": [
            "usp:m49500",
            "vandf:4019477",
            "unii:NR7O1405Q9",
            "atc:R05CB05",
            "atc:V03AF01",
            "mmsl:41498",
            "mmsl:5057",
            "mmsl:d01411",
        ],
        "trade_names": ["Mesna", "Mesnex", "Mesna Nova Plus"],
    }
    return Therapy(**params)


@pytest.fixture(scope="module")
def beta_alanine():
    """Create beta-alanine drug fixture."""
    params = {
        "label": "beta-alanine",
        "concept_id": "rxcui:61",
        "aliases": [],
        "approval_ratings": ["rxnorm_prescribable"],
        "xrefs": [],
        "associated_with": ["vandf:4028377", "unii:11P2JDE17B"],
        "trade_names": [],
    }
    return Therapy(**params)


@pytest.fixture(scope="module")
def algestone():
    """Create algestone drug fixture."""
    params = {
        "label": "algestone",
        "concept_id": "rxcui:595",
        "aliases": [],
        "approval_ratings": [],
        "xrefs": [],
        "associated_with": [],
        "trade_names": [],
    }
    return Therapy(**params)


@pytest.fixture(scope="module")
def levothyroxine():
    """Create levothyroxine drug fixture."""
    params = {
        "label": "levothyroxine",
        "concept_id": "rxcui:10582",
        "aliases": [
            "3,5,3',5'-Tetraiodo-L-thyronine",
            "Thyroxine",
            "O-(4-Hydroxy-3,5-diiodophenyl)-3,5-diiodo-L-tyrosine",
            "L-T4",
            "LT4",
            "T4",
            "3,3',5,5'-Tetraiodo-L-thyronine",
            "L-Thyroxine",
            "O-(4-Hydroxy-3,5-diidophenyl)-3,5-diiodo-L-tyrosine",
            "Levothyroxin",
            "4-(4-Hydroxy-3,5-diiodophenoxy)-3,5-diiodo-L-phenylalanine",
        ],
        "approval_ratings": ["rxnorm_prescribable"],
        "xrefs": ["drugbank:DB00451"],
        "associated_with": [
            "vandf:4022126",
            "mmsl:d00278",
            "unii:Q51BO43MG4",
        ],
        # trade names length > 20
    }
    return Therapy(**params)


@pytest.fixture(scope="module")
def fluoxetine():
    """Create fluoxetine drug fixture."""
    params = {
        "label": "fluoxetine",
        "concept_id": "rxcui:4493",
        "aliases": [
            "FLUoxetine",
            "(+-)-N-Methyl-gamma-(4-(trifluoromethyl)phenoxy)" "benzenepropanamine",
            "(+-)-N-Methyl-3-phenyl-3-((alpha,alpha,alpha-trifluoro-"
            "P-tolyl)oxy)propylamine",
        ],
        "approval_ratings": ["rxnorm_prescribable"],
        "xrefs": ["drugbank:DB00472"],
        "associated_with": [
            "mmsl:17711",
            "vandf:4019761",
            "mmsl:d00236",
            "atc:N06AB03",
            "unii:01K63SUP8D",
        ],
        "trade_names": [
            "Prozac",
            "Sarafem",
            "Symbyax",
            "Rapiflux",
            "Selfemra",
        ],
    }
    return Therapy(**params)


@pytest.fixture(scope="module")
def fluoxetine_hydrochloride():
    """Create fluoxetine hydrochloride drug fixture."""
    params = {
        "label": "fluoxetine hydrochloride",
        "concept_id": "rxcui:227224",
        "aliases": ["FLUoxetine hydrochloride"],
        "approval_ratings": ["rxnorm_prescribable"],
        "xrefs": [],
        "associated_with": [
            "usp:m33780",
            "unii:I9W7N6B1KJ",
            "mmsl:28237",
            "mmsl:4746",
            "vandf:4019389",
            "mmsl:41730",
            "mmsl:37675",
        ],
        "trade_names": ["FLUoxetine HCl", "FLUoxetine Hydrochloride", "RECONCILE"],
    }
    return Therapy(**params)


@pytest.fixture(scope="module")
def phenobarbital():
    """Create phenobarbital fixture."""
    return Therapy(
        concept_id="rxcui:8134",
        label="phenobarbital",
        aliases=[
            "Phenobarbituric Acid",
            "Phenylethylbarbiturate",
            "5-ethyl-5-phenylpyrimidine-2,4,6(1H,3H,5H)-trione",
            "5-Phenyl-5-ethylbarbituric acid",
            "Phenylethylbarbituric Acid",
            "5-Ethyl-5-phenylbarbituric acid",
            "Phenobarbitone",
            "Phenobarbitol",
            "5-ethyl-5-phenyl-2,4,6(1H,3H,5H)-pyrimidinetrione",
            "PHENobarbital",
            "5-Ethyl-5-phenyl-pyrimidine-2,4,6-trione",
            "Phenylethylmalonylurea",
        ],
        xrefs=["drugbank:DB01174"],
        associated_with=[
            "mmsl:d00340",
            "usp:m63400",
            "mmsl:2390",
            "mmsl:5272",
            "vandf:4017422",
            "unii:YQE403BP4D",
            "atc:N03AA02",
        ],
        approval_ratings=["rxnorm_prescribable"],
        approval_year=[],
        has_indication=[],
    )


def test_bifidobacterium_infantis(bifidobacterium_infantis, rxnorm, compare_records):
    """Test that bifidobacterium_ nfantis drug normalizes to
    correct drug concept.
    """
    # Concept ID Match
    response = rxnorm.search("RxCUI:100213")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], bifidobacterium_infantis)

    # Label Match
    response = rxnorm.search(" Bifidobacterium infantis")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 1
    compare_records(response.records[0], bifidobacterium_infantis)

    response = rxnorm.search("bifidobacterium infantis")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 1
    compare_records(response.records[0], bifidobacterium_infantis)

    # Trade Name Match
    response = rxnorm.search("ALIGN")
    assert response.match_type == MatchType.TRADE_NAME
    assert len(response.records) == 1
    compare_records(response.records[0], bifidobacterium_infantis)

    response = rxnorm.search("evivo")
    assert response.match_type == MatchType.TRADE_NAME
    assert len(response.records) == 1
    compare_records(response.records[0], bifidobacterium_infantis)

    # Xref Match
    response = rxnorm.search("drugbank:DB14222")
    assert response.match_type == MatchType.XREF
    assert len(response.records) == 1
    compare_records(response.records[0], bifidobacterium_infantis)


def test_cisplatin(cisplatin, rxnorm, compare_records):
    """Test that cisplatin drug normalizes to correct drug concept."""
    # Concept ID Match
    response = rxnorm.search("RxCUI:2555")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], cisplatin)

    # Label Match
    response = rxnorm.search("CISPLATIN")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 1
    compare_records(response.records[0], cisplatin)

    # Alias Match
    response = rxnorm.search("Cis-DDP")
    assert response.match_type == MatchType.ALIAS
    assert len(response.records) == 1
    compare_records(response.records[0], cisplatin)

    # Trade Name Match
    response = rxnorm.search("platinol")
    assert response.match_type == MatchType.TRADE_NAME
    assert len(response.records) == 1
    compare_records(response.records[0], cisplatin)

    # Xref Match
    response = rxnorm.search("drugbank:DB12117")
    assert response.match_type == MatchType.XREF
    assert len(response.records) == 1
    compare_records(response.records[0], cisplatin)


def test_amiloride_hydrochloride(amiloride_hydrochloride, rxnorm, compare_records):
    """Test that amiloride_hydrochloride hydrochloride, anhydrous drug
    normalizes to correct drug concept.
    """
    # Concept ID Match
    response = rxnorm.search("RxcUI:142424")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], amiloride_hydrochloride)

    # Label Match
    response = rxnorm.search("amiloride hydrochloride")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 1
    compare_records(response.records[0], amiloride_hydrochloride)

    # Trade Name Match
    response = rxnorm.search("Midamor")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 1
    compare_records(response.records[0], amiloride_hydrochloride)


def test_amiloride(amiloride, rxnorm, compare_records):
    """Test that amiloride hydrochloride, anhydrous drug normalizes to
    correct drug concept.
    """
    # Concept ID Match
    response = rxnorm.search("RxcUI:644")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], amiloride)

    # Label Match
    response = rxnorm.search("amiloride")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 1
    compare_records(response.records[0], amiloride)

    # Alias Match
    response = rxnorm.search("Amyloride")
    assert response.match_type == MatchType.ALIAS
    assert len(response.records) == 1
    compare_records(response.records[0], amiloride)

    response = rxnorm.search(
        "3,5-diamino-N-carbamimidoyl-6-" "chloropyrazine-2-carboxamide"
    )
    assert response.match_type == MatchType.ALIAS
    assert len(response.records) == 1
    compare_records(response.records[0], amiloride)


def test_timolol(timolol, rxnorm, compare_records):
    """Test that timolol drug normalizes to correct drug concept."""
    # Concept ID Match
    response = rxnorm.search("RxcUI:10600")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], timolol)

    # Label Match
    response = rxnorm.search("timolol")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 1
    compare_records(response.records[0], timolol)

    # Alias Match
    response = rxnorm.search(
        "(S)-1-(tert-butylamino)-3-[(4-"
        "morpholin-4-yl-1,2,5-thiadiazol"
        "-3-yl)oxy]propan-2-ol"
    )
    assert response.match_type == MatchType.ALIAS
    assert len(response.records) == 1
    compare_records(response.records[0], timolol)

    # Trade Name Match
    response = rxnorm.search("Betimol")
    assert response.match_type == MatchType.TRADE_NAME
    assert len(response.records) == 1
    compare_records(response.records[0], timolol)


def test_lymphocyte(lymphocyte, rxnorm, compare_records):
    """Test that lymphocyte drug normalizes to correct drug concept."""
    # Concept ID Match
    response = rxnorm.search("RxCUI:1011")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], lymphocyte)

    # Label Match
    response = rxnorm.search("lymphocyte immune globulin, " "anti-thymocyte globulin")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 1
    compare_records(response.records[0], lymphocyte)

    # Trade Name Match
    response = rxnorm.search("Thymoglobulin")
    assert response.match_type == MatchType.TRADE_NAME
    assert len(response.records) == 2
    response.records.sort(key=lambda r: r.concept_id)
    compare_records(response.records[0], lymphocyte)

    response = rxnorm.search("ATGAM")
    assert response.match_type == MatchType.TRADE_NAME
    assert len(response.records) == 2
    response.records.sort(key=lambda r: r.concept_id)
    compare_records(response.records[0], lymphocyte)


def test_aspirin(aspirin, rxnorm, compare_records):
    """Test that aspirin drug normalizes to correct drug concept."""
    # Concept ID Match
    response = rxnorm.search("RxcUI:1191")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], aspirin)

    # (Trade Name) No Match
    response = rxnorm.search("Anacin")
    assert response.match_type == MatchType.NO_MATCH
    assert len(response.records) == 0


def test_mesnan(mesna, rxnorm, compare_records):
    """Test that mesna drug normalizes to correct drug concept."""
    # Concept ID Match
    response = rxnorm.search("RxCUI:44")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], mesna)

    # Label Match
    response = rxnorm.search("mesna")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 1
    compare_records(response.records[0], mesna)

    # Trade Name Match
    response = rxnorm.search("Mesnex")
    assert response.match_type == MatchType.TRADE_NAME
    assert len(response.records) == 1
    compare_records(response.records[0], mesna)


def test_beta_alanine(beta_alanine, rxnorm, compare_records):
    """Test that beta_alanine drug normalizes to correct drug concept."""
    # Concept ID Match
    response = rxnorm.search("RxCUI:61")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], beta_alanine)

    # Label Match
    response = rxnorm.search("beta-alanine")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 1
    compare_records(response.records[0], beta_alanine)


def test_algestone(algestone, rxnorm, compare_records):
    """Test that algestone drug normalizes to correct drug concept."""
    # Concept ID Match
    response = rxnorm.search("RxCUI:595")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], algestone)

    # Label Match
    response = rxnorm.search("algestone")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 1
    compare_records(response.records[0], algestone)


def test_levothyroxine(levothyroxine, rxnorm, compare_records):
    """Test that levothyroxine drug normalizes to correct drug concept."""
    # Concept ID Match
    response = rxnorm.search("RxCUI:10582")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], levothyroxine)

    # Label Match
    response = rxnorm.search("levothyroxine")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 1
    compare_records(response.records[0], levothyroxine)

    # Alias Match
    response = rxnorm.search("LT4")
    assert response.match_type == MatchType.ALIAS
    assert len(response.records) == 1
    compare_records(response.records[0], levothyroxine)

    # trade name matches fail because they exceed the max # limit
    response = rxnorm.search("Unithroid")
    assert response.match_type == MatchType.NO_MATCH

    response = rxnorm.search("Euthyrox")
    assert response.match_type == MatchType.NO_MATCH

    # Xref Match
    response = rxnorm.search("DRUGBANK:DB00451")
    assert response.match_type == MatchType.XREF
    assert len(response.records) == 1
    compare_records(response.records[0], levothyroxine)


def test_fluoxetine(fluoxetine, rxnorm, compare_records):
    """Test that fluoxetine drug normalizes to correct drug concept."""
    # Concept ID Match
    response = rxnorm.search("RxCUI:4493")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], fluoxetine)

    # Label Match
    response = rxnorm.search("fluoxetine")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 1
    compare_records(response.records[0], fluoxetine)


def test_fluoxetine_hydrochloride(fluoxetine_hydrochloride, rxnorm, compare_records):
    """Test that fluoxetine_hydrochloride drug normalizes to correct drug
    concept.
    """
    # Concept ID Match
    response = rxnorm.search("RxCUI:227224")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], fluoxetine_hydrochloride)

    # Label Match
    response = rxnorm.search("fluoxetine hydrochloride")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 1
    compare_records(response.records[0], fluoxetine_hydrochloride)


def test_no_match(rxnorm):
    """Test that a term normalizes to correct drug concept as a NO match."""
    # Misspelled name
    response = rxnorm.search("17-hydroxycorticosteroi")
    assert response.match_type == MatchType.NO_MATCH
    assert len(response.records) == 0

    # Not storing foreign synonyms
    response = rxnorm.search("cisplatino")
    assert response.match_type == MatchType.NO_MATCH
    assert len(response.records) == 0

    # Wrong Namespace
    response = rxnorm.search("rxnorm:3")
    assert response.match_type == MatchType.NO_MATCH

    # Test white space in between id
    response = rxnorm.search("rxcui: 3")
    assert response.match_type == MatchType.NO_MATCH

    # Should not store brand name concepts as identity record
    response = rxnorm.search("rxcui:202433")
    assert response.match_type == MatchType.NO_MATCH

    # Test empty query
    response = rxnorm.search("")
    assert response.match_type == MatchType.NO_MATCH
    assert len(response.records) == 0


def test_brand_name_to_concept(rxnorm):
    """Test that brand names are correctly linked to identity concept."""
    assert (
        rxnorm._query_handler.db.get_rxnorm_id_by_brand("rxcui:1041527") == "rxcui:161"
    )
    assert rxnorm._query_handler.db.get_rxnorm_id_by_brand("rxcui:218330") == "rxcui:44"


def test_xref_lookup(rxnorm, bifidobacterium_infantis, amiloride, compare_records):
    """Test that xref lookup resolves to correct concept."""
    response = rxnorm.search("mmsl:d07347")
    assert response.match_type == MatchType.ASSOCIATED_WITH
    assert len(response.records) == 1
    compare_records(response.records[0], bifidobacterium_infantis)

    response = rxnorm.search("atc:C03DB01")
    assert response.match_type == MatchType.ASSOCIATED_WITH
    assert len(response.records) == 1
    compare_records(response.records[0], amiloride)


def test_phenobarbital(phenobarbital, rxnorm, compare_records):
    """Test phenobarbital search."""
    # Concept ID Match
    response = rxnorm.search("rxcui:8134")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], phenobarbital)

    # Label Match
    response = rxnorm.search(" phenobarbital")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 1
    compare_records(response.records[0], phenobarbital)

    # Xref Match
    response = rxnorm.search("drugbank:DB01174")
    assert response.match_type == MatchType.XREF
    assert len(response.records) == 1
    compare_records(response.records[0], phenobarbital)


def test_meta_info(rxnorm):
    """Test that the meta field is correct."""
    response = rxnorm.search("RxNorm")
    assert response.source_meta_.data_license == "UMLS Metathesaurus"
    assert (
        response.source_meta_.data_license_url
        == "https://www.nlm.nih.gov/research/umls/rxnorm/docs/termsofservice.html"
    )
    assert isodate.parse_date(response.source_meta_.version)
    assert response.source_meta_.data_url.startswith(
        "https://download.nlm.nih.gov/umls/kss/rxnorm/"
    )
    assert not response.source_meta_.rdp_url
    assert response.source_meta_.data_license_attributes == {
        "non_commercial": False,
        "share_alike": False,
        "attribution": True,
    }



"""ETL methods for NCIt source"""

import logging

import owlready2 as owl
from owlready2.entity import ThingClass
from tqdm import tqdm

from therapy.etl.base import Base
from therapy.schemas import NamespacePrefix, RecordParams, SourceMeta, SourceName

_logger = logging.getLogger(__name__)


class NCIt(Base):
    """Class for NCIt ETL methods.

    Extracts:
     * NCIt classes with semantic_type "Pharmacologic Substance" but not
       Retired_Concept
     * NCIt classes that are subclasses of C1909 (Pharmacologic Substance)
    """

    def _get_desc_nodes(
        self, node: ThingClass, uq_nodes: set[ThingClass]
    ) -> set[ThingClass]:
        """Create set of unique subclasses of node parameter.
        Should be originally called on ncit:C1909: Pharmacologic Substance.

        :param owlready2.entity.ThingClass node: concept node to either retrieve
            descendants of, or to normalize and add to DB
        :param Set[owlready2.entity.ThingClass] uq_nodes: set of unique class nodes
            found so far from recursive tree exploration
        :return: the uq_nodes set, updated with any class nodes found from recursive
            exploration of this branch of the class tree
        :rtype: Set[owlready2.entity.ThingClass]
        """
        children = node.descendants()
        if children:
            for child_node in children:
                if child_node is not node:
                    uq_nodes.add(child_node)
                    self._get_desc_nodes(child_node, uq_nodes)
        return uq_nodes

    def _get_typed_nodes(
        self, uq_nodes: set[ThingClass], ncit: owl.namespace.Ontology
    ) -> set[ThingClass]:
        """Get all nodes with semantic_type Pharmacologic Substance

        :param Set[owlready2.entity.ThingClass] uq_nodes: set of unique class nodes
            found so far.
        :param owl.namespace.Ontology ncit: owlready2 Ontology instance for NCIt.
        :return: uq_nodes, with the addition of all classes found to have semantic_type
            Pharmacologic Substance and not of type Retired_Concept
        :rtype: Set[owlready2.entity.ThingClass]
        """
        graph = owl.default_world.as_rdflib_graph()

        query_str = """SELECT ?x WHERE {
            ?x <http://ncicb.nci.nih.gov/xml/owl/EVS/Thesaurus.owl#P106>
            \"Pharmacologic Substance\"
        }
        """
        typed_results = set(graph.query(query_str))

        retired_query_str = """SELECT ?x WHERE {
            ?x <http://ncicb.nci.nih.gov/xml/owl/EVS/Thesaurus.owl#P310>
            \"Retired_Concept\"
        }
        """
        retired_results = set(graph.query(retired_query_str))

        typed_results = {r for r in (typed_results - retired_results) if r is not None}

        for result in typed_results:
            # parse result as URI and get ThingClass object back from NCIt
            class_object = ncit[result[0].toPython().split("#")[1]]
            uq_nodes.add(class_object)
        return uq_nodes

    def _transform_data(self) -> None:
        """Get data from file and construct objects for loading"""
        ncit = owl.get_ontology(self._data_file.absolute().as_uri())  # type: ignore
        ncit.load()
        uq_nodes = {ncit.C49236}  # add Therapeutic Procedure
        uq_nodes = self._get_desc_nodes(ncit.C1909, uq_nodes)
        uq_nodes = self._get_typed_nodes(uq_nodes, ncit)
        for node in tqdm(
            uq_nodes,
            ncols=80,
            disable=self._silent,
        ):
            concept_id = f"{NamespacePrefix.NCIT.value}:{node.name}"
            label = node.P108.first() if node.P108 else None
            aliases = node.P90.copy()  # prevent CallbackList error
            if label and aliases and label in aliases:
                aliases.remove(label)

            xrefs = []
            associated_with = []
            if node.P207:
                associated_with.append(
                    f"{NamespacePrefix.UMLS.value}:" f"{node.P207.first()}"
                )
            if node.P210:
                xrefs.append(
                    f"{NamespacePrefix.CASREGISTRY.value}:" f"{node.P210.first()}"
                )
            if node.P319:
                associated_with.append(
                    f"{NamespacePrefix.UNII.value}:" f"{node.P319.first()}"
                )
            if node.P368:
                iri = node.P368.first()
                if ":" in iri:
                    iri = iri.split(":")[1]
                associated_with.append(f"{NamespacePrefix.CHEBI.value}:{iri}")
            params: RecordParams = {  # type: ignore
                "concept_id": concept_id,
                "label": label,
                "aliases": aliases,
                "xrefs": xrefs,
                "associated_with": associated_with,
            }
            self._load_therapy(params)

    def _load_meta(self) -> None:
        """Load metadata"""
        metadata = SourceMeta(
            data_license="CC BY 4.0",
            data_license_url="https://creativecommons.org/licenses/by/4.0/legalcode",
            version=self._version,
            data_url="https://evs.nci.nih.gov/ftp1/NCI_Thesaurus/",
            rdp_url="http://reusabledata.org/ncit.html",
            data_license_attributes={
                "non_commercial": False,
                "share_alike": False,
                "attribution": True,
            },
        )
        self.database.add_source_metadata(SourceName.NCIT, metadata)



"""Provide functions and utilities related to application logging."""

import logging


def _quiet_upstream_libs() -> None:
    """Turn off debug logging for chatty upstream library loggers."""
    for lib in (
        "boto3",
        "botocore",
        "urllib3",
    ):
        logging.getLogger(lib).setLevel(logging.INFO)


def configure_logs(
    log_file: str | None = None,
    log_level: int = logging.DEBUG,
    quiet_upstream: bool = True,
) -> None:
    """Configure logging.

    :param log_filename: location to put log file at
    :param log_level: global log level to set
    :param quiet_upstream: if True, turn off debug logging for a selection of libraries
    """
    if log_file is None:
        log_file = "therapy.log"
    if quiet_upstream:
        _quiet_upstream_libs()
    logging.basicConfig(
        filename=log_file,
        format="[%(asctime)s] - %(name)s - %(levelname)s : %(message)s",
    )
    logger = logging.getLogger(__package__)
    logger.setLevel(log_level)



"""Construct test data for ChEMBL source."""

import contextlib
import sqlite3
from pathlib import Path

from therapy.database import create_db
from therapy.etl import ChEMBL

TEST_IDS = [
    "'CHEMBL11359'",
    "'CHEMBL2068237'",
    "'CHEMBL1628046'",
    "'CHEMBL267014'",
    "'CHEMBL25'",
    "'CHEMBL843'",
    "'CHEMBL40'",
    "'CHEMBL121'",
    "'CHEMBL1006'",
]

ch = ChEMBL(create_db())
ch._extract_data(False)

TEST_DATA_DIR = Path(__file__).resolve().parents[1] / "data" / "chembl"
out_db_path = TEST_DATA_DIR / ch._data_file.name  # type: ignore

with contextlib.suppress(FileNotFoundError):
    out_db_path.unlink()
out_con = sqlite3.connect(out_db_path)
c = out_con.cursor()
c.execute(f"ATTACH DATABASE '{ch._data_file}' AS chembl")

schema_query = """
SELECT sql
FROM chembl.sqlite_master
WHERE
    type='table' AND
    name IN (
        'molecule_dictionary',
        'molecule_synonyms',
        'formulations',
        'drug_indication',
        'products'
    )
"""
schema = [r[0] for r in c.execute(schema_query).fetchall()]
for table_schema in schema:
    c.execute(table_schema)

md_query = f"""
INSERT INTO molecule_dictionary
SELECT *
FROM chembl.molecule_dictionary
WHERE chembl.molecule_dictionary.chembl_id IN (
    {",".join(TEST_IDS)}
)
"""
c.execute(md_query)

ms_query = """
INSERT INTO molecule_synonyms
SELECT cms.* FROM chembl.molecule_synonyms cms
INNER JOIN molecule_dictionary md
ON cms.molregno = md.molregno
"""
c.execute(ms_query)

f_query = """
INSERT INTO formulations
SELECT cf.* FROM chembl.formulations cf
INNER JOIN molecule_dictionary md
ON cf.molregno = md.molregno
"""
c.execute(f_query)

p_query = """
INSERT INTO products
SELECT p.* FROM chembl.products p
INNER JOIN formulations f
ON p.product_id = f.product_id
"""
c.execute(p_query)

d_query = """
INSERT INTO drug_indication
SELECT d.* FROM chembl.drug_indication d
INNER JOIN molecule_dictionary md
ON d.molregno = md.molregno
"""
c.execute(d_query)

out_con.commit()
out_con.close()



"""Provides a CLI util to make updates to normalizer database."""

import logging
from collections.abc import Collection
from timeit import default_timer as timer

import click
from disease.cli import _update_sources as update_disease_sources
from disease.database import create_db as create_disease_db
from disease.schemas import SourceName as DiseaseSourceName

from therapy import SOURCES
from therapy.database.database import (
    AbstractDatabase,
    DatabaseReadError,
    DatabaseWriteError,
    create_db,
)
from therapy.log import configure_logs
from therapy.schemas import SourceName

_logger = logging.getLogger(__name__)


@click.command()
@click.option("--db_url", help="URL endpoint for the application database.")
@click.option("--verbose", "-v", is_flag=True, help="Print result to console if set.")
def check_db(db_url: str, verbose: bool = False) -> None:
    """Perform basic checks on DB health and population. Exits with status code 1
    if DB schema is uninitialized or if critical tables appear to be empty.

    \f
    :param db_url: URL to normalizer database
    :param verbose: if true, print result to console
    """  # noqa: D301
    configure_logs()
    db = create_db(db_url, False)
    if not db.check_schema_initialized():
        if verbose:
            click.echo("Health check failed: DB schema uninitialized.")
        click.get_current_context().exit(1)
    if verbose:
        click.echo("DB health check successful: tables appear complete.")


def _delete_source(n: SourceName, db: AbstractDatabase) -> float:
    """Delete individual source data.

    :param n: name of source to delete
    :param db: database instance
    :return: time taken (in seconds) to delete
    """
    msg = f"Deleting {n.value}..."
    click.echo(f"\n{msg}")
    _logger.info(msg)
    start_delete = timer()
    db.delete_source(n)
    end_delete = timer()
    delete_time = end_delete - start_delete
    msg = f"Deleted {n.value} in {delete_time:.5f} seconds."
    click.echo(f"{msg}\n")
    _logger.info(msg)
    return delete_time


def _load_source(
    name: SourceName,
    db: AbstractDatabase,
    delete_time: float,
    processed_ids: list[str],
    use_existing: bool,
) -> None:
    """Load individual source data.

    :param n: name of source
    :param db: database instance
    :param delete_time: time taken (in seconds) to run deletion
    :param processed_ids: in-progress list of processed therapy IDs
    :param use_existing: if True, use most recent local data files instead of
        fetching from remote
    """
    msg = f"Loading {name.value}..."
    click.echo(msg)
    _logger.info(msg)
    start_load = timer()

    # used to get source class name from string
    try:
        from therapy.etl import (  # noqa: F401
            ChEMBL,
            ChemIDplus,
            DrugBank,
            DrugsAtFDA,
            EtlError,
            GuideToPHARMACOLOGY,
            HemOnc,
            NCIt,
            RxNorm,
            Wikidata,
        )
    except ModuleNotFoundError as e:
        click.echo(
            f"Encountered ModuleNotFoundError attempting to import {e.name}. Are ETL dependencies installed?"
        )
        click.get_current_context().exit()
    SourceClass = eval(name.value)  # noqa: N806 PGH001 S307

    source = SourceClass(database=db, silent=False)
    try:
        processed_ids += source.perform_etl(use_existing)
    except EtlError as e:
        _logger.error(e)
        click.echo(f"Encountered error while loading {name}: {e}.")
        click.get_current_context().exit()
    end_load = timer()
    load_time = end_load - start_load
    msg = f"Loaded {name.value} in {load_time:.5f} seconds."
    click.echo(msg)
    _logger.info(msg)
    msg = f"Total time for {name.value}: {(delete_time + load_time):.5f} seconds."
    click.echo(msg)
    _logger.info(msg)


def _delete_normalized_data(database: AbstractDatabase) -> None:
    """Delete normalized concepts

    :param database: DB instance
    """
    click.echo("\nDeleting normalized records...")
    start_delete = timer()
    try:
        database.delete_normalized_concepts()
    except (DatabaseReadError, DatabaseWriteError) as e:
        click.echo(f"Encountered exception during normalized data deletion: {e}")
    end_delete = timer()
    delete_time = end_delete - start_delete
    click.echo(f"Deleted normalized records in {delete_time:.5f} seconds.")


def _load_merge(db: AbstractDatabase, processed_ids: set[str]) -> None:
    """Load merged concepts

    :param db: database instance
    :param processed_ids: in-progress list of processed therapy IDs
    """
    start = timer()
    _delete_normalized_data(db)
    if not processed_ids:
        processed_ids = db.get_all_concept_ids()
    try:
        from therapy.etl.merge import Merge
    except ModuleNotFoundError as e:
        click.echo(
            f"Encountered ModuleNotFoundError attempting to import {e.name}. Are ETL dependencies installed?"
        )
        click.get_current_context().exit()

    merge = Merge(database=db, silent=False)
    click.echo("Constructing normalized records...")
    merge.create_merged_concepts(processed_ids)
    end = timer()
    click.echo(
        f"Merged concept generation completed in " f"{(end - start):.5f} seconds"
    )


def _update_normalizer(
    sources: Collection[SourceName],
    db: AbstractDatabase,
    update_merged: bool,
    use_existing: bool,
) -> None:
    """Update selected normalizer sources.

    :param sources: names of sources to update
    :param db: database instance
    :param update_merged: if true, retain processed records to use in updating merged
        records
    :param use_existing: if True, use most recent local version of source data instead of
        fetching from remote
    """
    processed_ids: list[str] = []
    for n in sources:
        delete_time = _delete_source(n, db)
        _load_source(n, db, delete_time, processed_ids, use_existing)

    if update_merged:
        _load_merge(db, set(processed_ids))


def _ensure_diseases_updated(from_local: bool) -> None:
    """Check disease DB. If it appears unpopulate, run a full update.

    :param from_local: if True, update from most recent locally available data
    """
    disease_db = create_disease_db()
    if (
        not disease_db.check_schema_initialized()
        or not disease_db.check_tables_populated()
    ):
        update_disease_sources(list(DiseaseSourceName), disease_db, True, from_local)


@click.command()
@click.option("--sources", help="The source(s) you wish to update separated by spaces.")
@click.option("--aws_instance", is_flag=True, help="Using AWS DynamodDB instance.")
@click.option("--db_url", help="URL endpoint for the application database.")
@click.option("--update_all", is_flag=True, help="Update all normalizer sources.")
@click.option(
    "--update_merged",
    is_flag=True,
    help="Update concepts for normalize endpoint from accepted sources.",
)
@click.option(
    "--use_existing",
    is_flag=True,
    default=False,
    help="Use most recent local source data instead of fetching latest version",
)
def update_normalizer_db(
    sources: str,
    aws_instance: bool,
    db_url: str,
    update_all: bool,
    update_merged: bool,
    use_existing: bool,
) -> None:
    """Update selected normalizer source(s) in the therapy database. For example, the
    following command will update RxNorm and Wikidata data, using a database connection at port 8001:

    % therapy_norm_update --sources="rxnorm wikidata" --db_url=http://localhost:8001

    \f
    :param sources: names of sources to update, comma-separated
    :param aws_instance: if true, use cloud instance
    :param db_url: URI pointing to database
    :param update_all: if true, update all sources (ignore `normalizer` parameter)
    :param update_merged: if true, update normalized records
    :param use_existing: if True, use most recent local data instead of fetching latest version
    """  # noqa: D301
    configure_logs()
    db = create_db(db_url, aws_instance)

    if update_all:
        _ensure_diseases_updated(use_existing)
        _update_normalizer(list(SourceName), db, update_merged, use_existing)
    elif not sources:
        if update_merged:
            _load_merge(db, set())
        else:
            ctx = click.get_current_context()
            click.echo(
                "Must either enter 1 or more sources, or use `--update_all` parameter"
            )
            click.echo(ctx.get_help())
            ctx.exit()
    else:
        sources_split = sources.lower().split()

        if len(sources_split) == 0:
            msg = "Must enter 1 or more source names to update"
            raise Exception(msg)

        non_sources = set(sources_split) - set(SOURCES)

        if len(non_sources) != 0:
            msg = f"Not valid source(s): {non_sources}"
            raise Exception(msg)

        parsed_source_names = {SourceName(SOURCES[s]) for s in sources_split}
        if (
            SourceName.CHEMBL in parsed_source_names
            or SourceName.HEMONC in parsed_source_names
        ):
            _ensure_diseases_updated(use_existing)
        _update_normalizer(parsed_source_names, db, update_merged, use_existing)


if __name__ == "__main__":
    update_normalizer_db()



"""Test that the therapy normalizer works as intended for the DrugBank
source.
"""

import re

import pytest

from therapy.etl.drugbank import DrugBank
from therapy.schemas import MatchType, Therapy


@pytest.fixture(scope="module")
def drugbank(test_source):
    """Provide test DrugBank query endpoint"""
    return test_source(DrugBank)


@pytest.fixture(scope="module")
def cisplatin():
    """Create a cisplatin drug fixture."""
    params = {
        "label": "Cisplatin",
        "concept_id": "drugbank:DB00515",
        "aliases": [
            "CDDP",
            "Cis-DDP",
            "cis-diamminedichloroplatinum(II)",
            "APRD00359",
            "cisplatino",
        ],
        "xrefs": [
            "chemidplus:15663-27-1",
        ],
        "trade_names": [],
        "associated_with": ["unii:Q20Q21Q62J", "inchikey:LXZZYRPGZAFOLE-UHFFFAOYSA-L"],
    }
    return Therapy(**params)


@pytest.fixture(scope="module")
def bentiromide():
    """Create a bentiromide drug fixture."""
    params = {
        "label": "Bentiromide",
        "concept_id": "drugbank:DB00522",
        "aliases": [
            "APRD00818",
            "Bentiromido",
            "Bentiromidum",
            "BTPABA",
            "PFT",
        ],
        "xrefs": [
            "chemidplus:37106-97-1",
        ],
        "trade_names": [],
        "associated_with": ["unii:239IF5W61J", "inchikey:SPPTWHFVYKCNNK-FQEVSTJZSA-N"],
    }
    return Therapy(**params)


@pytest.fixture(scope="module")
def aloe_ferox_leaf():
    """Create aloe ferox leaf fixture. Record has >20 aliases, so ETL deletes
    all of them.
    """
    params = {
        "concept_id": "drugbank:DB14257",
        "label": "Aloe ferox leaf",
        "aliases": [],
        "xrefs": [],
        "trade_names": [],
        "associated_with": ["unii:0D145J8EME"],
    }
    return Therapy(**params)


def test_concept_id_match(
    drugbank, compare_response, cisplatin, bentiromide, aloe_ferox_leaf
):
    """Test that concept ID query resolves to correct record."""
    response = drugbank.search("drugbank:DB00515")
    compare_response(response, MatchType.CONCEPT_ID, cisplatin)

    response = drugbank.search("DB00515")
    compare_response(response, MatchType.CONCEPT_ID, cisplatin)

    response = drugbank.search("drugbank:db00515")
    compare_response(response, MatchType.CONCEPT_ID, cisplatin)

    response = drugbank.search("Drugbank:db00515")
    compare_response(response, MatchType.CONCEPT_ID, cisplatin)

    response = drugbank.search("druGBank:DB00515")
    compare_response(response, MatchType.CONCEPT_ID, cisplatin)

    response = drugbank.search("drugbank:DB00522")
    compare_response(response, MatchType.CONCEPT_ID, bentiromide)

    response = drugbank.search("DB00522")
    compare_response(response, MatchType.CONCEPT_ID, bentiromide)

    response = drugbank.search("drugbank:db00522")
    compare_response(response, MatchType.CONCEPT_ID, bentiromide)

    response = drugbank.search("Drugbank:db00522")
    compare_response(response, MatchType.CONCEPT_ID, bentiromide)

    response = drugbank.search("druGBank:DB00522")
    compare_response(response, MatchType.CONCEPT_ID, bentiromide)

    response = drugbank.search("drugbank:DB14257")
    compare_response(response, MatchType.CONCEPT_ID, aloe_ferox_leaf)


def test_label_match(
    drugbank, compare_response, cisplatin, bentiromide, aloe_ferox_leaf
):
    """Test that label query resolves to correct record."""
    response = drugbank.search("cisplatin")
    compare_response(response, MatchType.LABEL, cisplatin)

    response = drugbank.search("cisplatin")
    compare_response(response, MatchType.LABEL, cisplatin)

    response = drugbank.search("Bentiromide")
    compare_response(response, MatchType.LABEL, bentiromide)

    response = drugbank.search("bentiromide")
    compare_response(response, MatchType.LABEL, bentiromide)

    response = drugbank.search("aloe ferox leaf")
    compare_response(response, MatchType.LABEL, aloe_ferox_leaf)


def test_alias_match(drugbank, compare_response, cisplatin, bentiromide):
    """Test that alias query resolves to correct record."""
    response = drugbank.search("CISPLATINO")
    compare_response(response, MatchType.ALIAS, cisplatin)

    response = drugbank.search("Cis-ddp")
    compare_response(response, MatchType.ALIAS, cisplatin)

    response = drugbank.search("APRD00818")
    compare_response(response, MatchType.ALIAS, bentiromide)

    response = drugbank.search("PFT")
    compare_response(response, MatchType.ALIAS, bentiromide)

    # verify aliases > 20 not stored
    response = drugbank.search("Aloe Capensis")
    assert response.match_type == MatchType.NO_MATCH

    response = drugbank.search("Aloe Ferox Juice")
    assert response.match_type == MatchType.NO_MATCH


def test_xref_match(drugbank, compare_response, cisplatin, bentiromide):
    """Test that xref query resolves to correct record."""
    response = drugbank.search("chemidplus:15663-27-1")
    compare_response(response, MatchType.XREF, cisplatin)

    response = drugbank.search("chemidplus:37106-97-1")
    compare_response(response, MatchType.XREF, bentiromide)


def test_assoc_with_match(
    drugbank, compare_response, cisplatin, bentiromide, aloe_ferox_leaf
):
    """Test that associated_with query resolves to correct record."""
    response = drugbank.search("inchikey:lxzzyrpgzafole-uhfffaoysa-l")
    compare_response(response, MatchType.ASSOCIATED_WITH, cisplatin)

    response = drugbank.search("unii:239if5w61j")
    compare_response(response, MatchType.ASSOCIATED_WITH, bentiromide)

    response = drugbank.search("UNII:0D145J8EME")
    compare_response(response, MatchType.ASSOCIATED_WITH, aloe_ferox_leaf)


def test_no_match(drugbank):
    """Test that a term normalizes to correct drug concept as a NO match."""
    response = drugbank.search("lepirudi")
    assert response.match_type == MatchType.NO_MATCH
    assert len(response.records) == 0

    # Test white space in between id
    response = drugbank.search("DB 00001")
    assert response.match_type == MatchType.NO_MATCH

    # Test empty query
    response = drugbank.search("")
    assert response.match_type == MatchType.NO_MATCH
    assert len(response.records) == 0


def test_meta_info(drugbank):
    """Test that the meta field is correct."""
    response = drugbank.search("search")
    assert response.source_meta_.data_license == "CC0 1.0"
    assert (
        response.source_meta_.data_license_url
        == "https://creativecommons.org/publicdomain/zero/1.0/"
    )
    assert re.match(r"[0-9]+\.[0-9]+\.[0-9]", response.source_meta_.version)
    assert (
        response.source_meta_.data_url
        == "https://go.drugbank.com/releases/latest#open-data"
    )
    assert response.source_meta_.rdp_url == "http://reusabledata.org/drugbank.html"
    assert response.source_meta_.data_license_attributes == {
        "non_commercial": False,
        "share_alike": False,
        "attribution": False,
    }



"""Defines the Wikidata ETL methods."""

import datetime
import json
import os
from pathlib import Path
from typing import Any

from tqdm import tqdm
from wags_tails import CustomData, DataSource
from wags_tails.utils.versioning import DATE_VERSION_PATTERN
from wikibaseintegrator.wbi_config import config as wbi_config
from wikibaseintegrator.wbi_helpers import execute_sparql_query

from therapy import XREF_SOURCES
from therapy.etl.base import Base, EtlError
from therapy.schemas import NamespacePrefix, RecordParams, SourceMeta, SourceName

wbi_config["USER_AGENT"] = (
    os.environ.get("WIKIBASE_USER_AGENT")
    or "thera_py (https://github.com/cancervariants/therapy-normalization)"
)

# Translate Wikidata keys to standardized namespaces
NAMESPACES = {
    "casRegistry": NamespacePrefix.CASREGISTRY.value,
    "ChemIDplus": NamespacePrefix.CHEMIDPLUS.value,
    "pubchemCompound": NamespacePrefix.PUBCHEMCOMPOUND.value,
    "pubchemSubstance": NamespacePrefix.PUBCHEMSUBSTANCE.value,
    "chembl": NamespacePrefix.CHEMBL.value,
    "rxnorm": NamespacePrefix.RXNORM.value,
    "drugbank": NamespacePrefix.DRUGBANK.value,
    "wikidata": NamespacePrefix.WIKIDATA.value,
    "guideToPharmacology": NamespacePrefix.GUIDETOPHARMACOLOGY.value,
}

# Provide standard concept ID prefixes
ID_PREFIXES = {
    "wikidata": "Q",
    "chembl": "CHEMBL",
    "drugbank": "DB",
    "ncit": "C",
}

SPARQL_QUERY = """
SELECT
  ?item ?itemLabel ?casRegistry ?pubchemCompound ?pubchemSubstance ?chembl ?rxnorm
  ?drugbank ?guideToPharmacology
  (GROUP_CONCAT(DISTINCT ?alias; separator=";;") AS ?aliases)
WHERE {
  { ?item (wdt:P31/(wdt:P279*)) wd:Q12140. }
  UNION
  { ?item (wdt:P366/(wdt:P279*)) wd:Q12140. }
  UNION
  { ?item (wdt:P31/(wdt:P279*)) wd:Q35456. }
  UNION
  { ?item (wdt:P2868/(wdt:P279*)) wd:Q12187. }
  OPTIONAL {
    ?item skos:altLabel ?alias.
    FILTER((LANG(?alias)) = "en")
  }
  OPTIONAL {
    ?item p:P231 ?wds1.
    ?wds1 ps:P231 ?casRegistry.
  }
  OPTIONAL {
    ?item p:P662 ?wds2.
    ?wds2 ps:P662 ?pubchemCompound.
  }
  OPTIONAL {
    ?item p:P2153 ?wds3.
    ?wds3 ps:P2153 ?pubchemSubstance.
  }
  OPTIONAL {
    ?item p:P592 ?wds4.
    ?wds4 ps:P592 ?chembl.
  }
  OPTIONAL {
    ?item p:P3345 ?wds5.
    ?wds5 ps:P3345 ?rxnorm.
  }
  OPTIONAL {
    ?item p:P715 ?wds6.
    ?wds6 ps:P715 ?drugbank.
  }
  OPTIONAL {
      ?item p:P595 ?wds7.
      ?wds7 ps:P595 ?guideToPharmacology
  }
  SERVICE wikibase:label { bd:serviceParam wikibase:language "[AUTO_LANGUAGE],en". }
}
GROUP BY ?item ?itemLabel ?casRegistry ?pubchemCompound ?pubchemSubstance ?chembl
  ?rxnorm ?drugbank ?guideToPharmacology
"""


class Wikidata(Base):
    """Class for Wikidata ETL methods."""

    @staticmethod
    def _download_data(version: str, file: Path) -> None:  # noqa: ARG004
        """Download latest Wikidata source dump.

        :param version: not used by this method
        :param file: location to save data dump at
        :raise EtlError: if SPARQL query fails
        """
        medicine_query_results = execute_sparql_query(SPARQL_QUERY)
        if medicine_query_results is None:
            msg = "Wikidata medicine SPARQL query failed"
            raise EtlError(msg)
        results = medicine_query_results["results"]["bindings"]

        transformed_data = []
        for item in results:
            params: RecordParams = {}
            for attr in item:
                params[attr] = item[attr]["value"]
            transformed_data.append(params)
        with file.open("w+") as f:
            json.dump(transformed_data, f)

    @staticmethod
    def _get_latest_version() -> str:
        """Wikidata is constantly, immediately updated, so source data has no strict
        versioning. We use the current date as a pragmatic way to indicate the version.

        :return: formatted string for the current date
        """
        return datetime.datetime.now(tz=datetime.UTC).strftime(DATE_VERSION_PATTERN)

    def _get_data_handler(self, data_path: Path | None = None) -> DataSource:
        """Construct data handler instance for source. Overwrites base class method
        to use custom data handler instead.

        :param data_path: location of data storage
        :return: instance of wags_tails.DataSource to manage source file(s)
        """
        return CustomData(
            "wikidata",
            "json",
            self._get_latest_version,
            self._download_data,
            data_dir=data_path,
            silent=self._silent,
        )

    def _load_meta(self) -> None:
        """Add Wikidata metadata."""
        metadata = SourceMeta(
            data_license="CC0 1.0",
            data_license_url="https://creativecommons.org/publicdomain/zero/1.0/",
            version=self._version,
            data_url=None,
            rdp_url=None,
            data_license_attributes={
                "non_commercial": False,
                "share_alike": False,
                "attribution": False,
            },
        )
        self.database.add_source_metadata(SourceName.WIKIDATA, metadata)

    def _transform_data(self) -> None:
        """Transform the Wikidata source data."""
        with self._data_file.open() as f:
            records: dict = json.load(f)

            items: dict[str, Any] = {}

            for record in records:
                record_id = record["item"].split("/")[-1]
                concept_id = f"{NamespacePrefix.WIKIDATA.value}:{record_id}"
                if concept_id not in items:
                    item: dict[str, Any] = {"concept_id": concept_id}

                    xrefs = []
                    associated_with = []
                    for key in NAMESPACES:
                        if key in record:
                            ref = record[key]

                            if key.upper() == "CASREGISTRY":
                                key = SourceName.CHEMIDPLUS.value

                            if key.upper() in XREF_SOURCES:
                                if key != "chembl":
                                    prefix = ID_PREFIXES.get(key.lower(), "")
                                    fmted_xref = f"{NAMESPACES[key]}:{prefix}{ref}"
                                else:
                                    fmted_xref = f"{NAMESPACES[key]}:{ref}"
                                xrefs.append(fmted_xref)
                            else:
                                fmted_assoc = f"{NAMESPACES[key]}:{ref}"
                                associated_with.append(fmted_assoc)
                    item["xrefs"] = xrefs
                    item["associated_with"] = associated_with
                    if "itemLabel" in record:
                        item["label"] = record["itemLabel"]
                    items[concept_id] = item
                if "aliases" in record:
                    if "aliases" in items[concept_id]:
                        items[concept_id]["aliases"] += record["aliases"].split(";;")
                    else:
                        items[concept_id]["aliases"] = record["aliases"].split(";;")
        for item in tqdm(items.values(), ncols=80, disable=self._silent):
            self._load_therapy(item)



"""Provide ETL methods for HemOnc.org data."""

import csv
import logging

from tqdm import tqdm
from wags_tails.hemonc import HemOncPaths

from therapy.etl.base import DiseaseIndicationBase
from therapy.schemas import (
    ApprovalRating,
    NamespacePrefix,
    RecordParams,
    SourceMeta,
)

_logger = logging.getLogger(__name__)


class HemOnc(DiseaseIndicationBase):
    """Class for HemOnc.org ETL methods."""

    def _extract_data(self, use_existing: bool) -> None:
        """Acquire source data.

        This method is responsible for initializing an instance of a data handler and
        setting ``self._data_files`` and ``self._version``.
        """
        data_files, self._version = self._data_source.get_latest(
            from_local=use_existing
        )
        self._data_files: HemOncPaths = data_files  # type: ignore

    def _load_meta(self) -> None:
        """Add HemOnc metadata."""
        meta = {
            "data_license": "CC BY 4.0",
            "data_license_url": "https://creativecommons.org/licenses/by/4.0/legalcode",
            "version": self._version,
            "data_url": "https://dataverse.harvard.edu/dataset.xhtml?persistentId=doi:10.7910/DVN/9CY9C6",
            "rdp_url": None,
            "data_license_attributes": {
                "non_commercial": False,
                "share_alike": False,
                "attribution": True,
            },
        }
        self.database.add_source_metadata(self._name, SourceMeta(**meta))

    def _get_concepts(self) -> tuple[dict, dict, dict]:
        """Get therapy, brand name, and disease concepts from concepts file.
        :return: Tuple of dicts mapping ID to object for each type of concept
        """
        therapies: RecordParams = {}  # hemonc id -> record
        brand_names: dict[str, str] = {}  # hemonc id -> brand name
        conditions: dict[str, str] = {}  # hemonc id -> condition name

        with self._data_files.concepts.open() as concepts_file:
            concepts_reader = csv.reader(concepts_file)
            next(concepts_reader)  # skip header
            for row in concepts_reader:
                if row[6]:
                    continue  # skip if deprecated/invalid

                row_type = row[2]
                if row_type == "Component":
                    concept_id = f"{NamespacePrefix.HEMONC.value}:{row[3]}"
                    therapies[row[3]] = {
                        "concept_id": concept_id,
                        "label": row[0],
                        "trade_names": [],
                        "aliases": [],
                        "xrefs": [],
                    }
                elif row_type == "Brand Name":
                    brand_names[row[3]] = row[0]
                elif row_type == "Condition":
                    conditions[row[3]] = row[0]

        return therapies, brand_names, conditions

    @staticmethod
    def _id_to_yr(hemonc_id: str) -> str:
        """Get year from HemOnc ID corresponding to year concept.
        :param str hemonc_id: HemOnc ID to get year for
        :return: str representing year. Raises TypeError if HemOnc ID not valid.
        """
        id_int = int(hemonc_id)
        if id_int == 780:
            return "9999"
        if id_int == 48349:
            return "2020"
        if id_int == 5963:
            return "2021"
        if id_int < 699 or id_int > 780:
            msg = "ID not a valid HemOnc year concept"
            raise TypeError(msg)
        return str(id_int + 1240)

    def _get_rels(self, therapies: dict, brand_names: dict, conditions: dict) -> dict:
        """Gather relations to provide associations between therapies, brand names,
        and conditions.

        :param dict therapies: mapping from IDs to therapy concepts
        :param dict brand_names: mapping from IDs to therapy brand names
        :param dict conditions: mapping from IDs to disease conditions
        :return: therapies dict updated with brand names and conditions
        """
        with self._data_files.rels.open() as rels_file:
            rels_reader = csv.reader(rels_file)
            next(rels_reader)  # skip header

            for row in rels_reader:
                rel_type = row[4]
                hemonc_id = row[0]
                record = therapies.get(hemonc_id)

                if record is None:
                    continue  # skip non-drug items

                if rel_type == "Maps to":
                    if " and " in record.get("label", ""):
                        # xref from the HemOnc combo treatment to individual drugs
                        # see https://github.com/cancervariants/therapy-normalization/issues/417
                        continue
                    src_raw = row[3]
                    if src_raw == "RxNorm":
                        xref = f"{NamespacePrefix.RXNORM.value}:{row[1]}"
                        record["xrefs"].append(xref)
                    elif src_raw == "RxNorm Extension":
                        continue  # skip
                    else:
                        _logger.warning("Unrecognized `Maps To` source: %s", src_raw)

                elif rel_type == "Has brand name":
                    try:
                        record["trade_names"].append(brand_names[row[1]])
                    except KeyError:
                        _logger.warning(
                            "Unrecognized brand name ID (%s) for HemOnc concept %s",
                            row[1],
                            row[0],
                        )
                        continue

                elif rel_type == "Was FDA approved yr":
                    try:
                        year = self._id_to_yr(row[1])
                    except TypeError:
                        _logger.error(
                            "Failed parse of FDA approval year ID %s for HemOnc ID %s",
                            row[1],
                            row[0],
                        )
                        continue
                    if year == "9999":
                        _logger.warning(
                            "HemOnc ID %s has FDA approval year 9999", row[0]
                        )
                    record["approval_ratings"] = [ApprovalRating.HEMONC_APPROVED.value]
                    if "approval_year" in record:
                        record["approval_year"].append(year)
                    else:
                        record["approval_year"] = [year]

                elif rel_type == "Has FDA indication":
                    try:
                        label = conditions[row[1]]
                    except KeyError:
                        # concept is deprecated or otherwise unavailable
                        _logger.error(
                            "Unable to process relation with indication %s -- deprecated?",
                            row[0],
                        )
                        continue
                    norm_id = self._normalize_disease(label)
                    hemonc_concept_id = f"{NamespacePrefix.HEMONC.value}:{row[1]}"
                    indication = {
                        "disease_id": hemonc_concept_id,
                        "disease_label": label,
                        "normalized_disease_id": norm_id,
                        "supplemental_info": {"regulatory_body": "FDA"},
                    }
                    if "has_indication" in record:
                        record["has_indication"].append(indication)
                    else:
                        record["has_indication"] = [indication]

        return therapies

    def _get_synonyms(self, therapies: dict) -> dict:
        """Gather synonym entries and associate with therapy concepts.

        :param dict therapies: mapping of IDs to therapy objects
        :return: therapies dict with synonyms added as aliases
        """
        with self._data_files.synonyms.open() as synonyms_file:
            synonyms_reader = csv.reader(synonyms_file)
            next(synonyms_reader)
            for row in synonyms_reader:
                therapy_code = row[1]
                if therapy_code in therapies:
                    therapy = therapies[therapy_code]
                    alias = row[0]
                    if alias != therapy.get("label"):
                        therapies[therapy_code]["aliases"].append(row[0])
            return therapies

    def _transform_data(self) -> None:
        """Prepare dataset for loading into normalizer database."""
        therapies, brand_names, conditions = self._get_concepts()
        therapies = self._get_rels(therapies, brand_names, conditions)
        therapies = self._get_synonyms(therapies)

        for therapy in tqdm(therapies.values(), ncols=80, disable=self._silent):
            self._load_therapy(therapy)



"""The VICC library for normalizing therapies."""

import re
from importlib.metadata import PackageNotFoundError, version
from pathlib import Path

APP_ROOT: Path = Path(__file__).resolve().parents[0]


try:
    __version__ = version("thera-py")
except PackageNotFoundError:
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError


from therapy.schemas import (  # noqa: E402
    NamespacePrefix,
    RefType,
    SourceName,
)

# map plural to singular form
# eg {"label": "label", "trade_names": "trade_name"}
# key is the field name in the record object, value is the item_type value
# in reference objects
ITEM_TYPES = {k.lower(): v.value for k, v in RefType.__members__.items()}

# Sources we import directly
SOURCES = {
    source.value.lower(): source.value for source in SourceName.__members__.values()
}

# use to fetch source name from schema based on concept id namespace
# e.g. {'chembl': 'ChEMBL'}
PREFIX_LOOKUP = {
    v.value: SourceName[k].value
    for k, v in NamespacePrefix.__members__.items()
    if k in SourceName.__members__
}

# Namespace LUI patterns. Use for namespace inference.
NAMESPACE_LUIS = (
    (re.compile(r"^CHEMBL\d+$", re.IGNORECASE), SourceName.CHEMBL.value),
    (re.compile(r"^\d+\-\d+\-\d+$", re.IGNORECASE), SourceName.CHEMIDPLUS.value),
    (re.compile(r"^(Q|P)\d+$", re.IGNORECASE), SourceName.WIKIDATA.value),
    (re.compile(r"^C\d+$", re.IGNORECASE), SourceName.NCIT.value),
    (re.compile(r"^DB\d{5}$", re.IGNORECASE), SourceName.DRUGBANK.value),
    (re.compile(r"^(A?NDA)(\d+)$", re.IGNORECASE), SourceName.DRUGSATFDA.value),
)

# Sources that we import directly
XREF_SOURCES = set(SourceName.__members__)

# Sources that are found in data from imported sources
ASSOC_WITH_SOURCES = set(NamespacePrefix.__members__) - XREF_SOURCES



"""Contains data models for representing VICC therapy records."""

from datetime import datetime
from enum import Enum, IntEnum
from typing import Any, Literal

from ga4gh.core.models import MappableConcept
from pydantic import BaseModel, ConfigDict, StrictBool, constr

from therapy import __version__

# Working structure for object in preparation for upload to DB
RecordParams = dict[str, list | set | str | dict[str, Any]]


class ApprovalRating(str, Enum):
    """Define string constraints for approval rating attribute. This reflects a drug's
    regulatory approval status as evaluated by an individual source. We opted to retain
    each individual source's rating as a distinct enum value after finding that some
    sources disagreed on the true value (either because of differences in scope like
    US vs EU/other regulatory arenas, old or conflicting data, or other reasons). Value
    descriptions are provided below from each listed source.

    ChEMBL: [http://chembl.blogspot.com/2023/03/what-is-max-phase-in-chembl.html]
     - CHEMBL_NULL: "preclinical compounds with bioactivity data"
     - CHEMBL_0_5: "A clinical candidate drug in Early Phase 1 Clinical Trials e.g.
    CITRULLINE MALATE (CHEMBL4297667) is under clinical investigation for coronary
     - CHEMBL_1: "A clinical candidate drug in Phase 1 Clinical Trials e.g.
    SALCAPROZATE SODIUM (CHEMBL2107027) is under clinical investigation for treatment
    of diabetes mellitus. Note that this category also includes a small number of
    trials that are defined by ClinicalTrials.gov as "Phase 1/Phase 2"."
    artery disease at Early Phase 1."
     - CHEMBL_2: "A clinical candidate drug in Phase 2 Clinical Trials e.g.
    NEVANIMIBE HYDROCHLORIDE (CHEMBL542103) is under clinical investigation for
    treatment of Cushing syndrome at Phase 2. Note that this category also includes a
    small number of trials that are defined by ClinicalTrials.gov as "Phase 2/Phase 3"."
     - CHEMBL_3: "A clinical candidate drug in Phase 3 Clinical Trials e.g.
    TEGOPRAZAN (CHEMBL4297583) is under clinical investigation for treatment of peptic
    ulcer at Phase 3, and also liver disease at Phase 1. "
     - CHEMBL_4: "A marketed drug e.g. AMINOPHYLLINE (CHEMBL1370561) is an FDA
    approved drug for treatment of asthma. "
    https://pubs.acs.org/doi/10.1021/acs.chemrestox.0c00296
     - note: "Unknown (-1)" should simply be left as None, i.e. not assigned an
    `ApprovalRating`.

    Drugs@FDA:
     - FDA_PRESCRIPTION: "A prescription drug product requires a doctor's authorization
    to purchase."
     - FDA_OTC: "FDA defines OTC drugs as safe and effective for use by the general
    public without a doctor's prescription."
     - FDA_DISCONTINUED: "approved products that have never been marketed, have been
    discontinued from marketing, are for military use, are for export only, or have had
    their approvals withdrawn for reasons other than safety or efficacy after being
    discontinued from marketing"
     - FDA_TENTATIVE: "If a generic drug product is ready for approval before the
    expiration of any patents or exclusivities accorded to the reference listed drug
    product, FDA issues a tentative approval letter to the applicant.  FDA delays final
    approval of the generic drug product until all patent or exclusivity issues have
    been resolved."
    https://www.fda.gov/drugs/drug-approvals-and-databases/drugsfda-glossary-terms

    HemOnc.org:
    - HEMONC_APPROVED: Inferred by us if "Was FDA Approved Yr" property is present
    (described as "Year of FDA approval")
    https://www.ncbi.nlm.nih.gov/pmc/articles/PMC6697579/

    Guide to Pharmacology:
    - GTOPDB_APPROVED: "Indicates pharmacologicaly active substances, specified by their
    INNs, that have been approved for clinical use by a regulatory agency, typically the
    FDA, EMA or in Japan. This classification persists regardless of whether the drug
    may later have been withdrawn or discontinued. (N.B. in some cases the information
    on approval status was obtained indirectly via databases such as Drugbank.)"
    - GTOPDB_WITHDRAWN: "No longer approved for its original clinical use, as notified
    by the FDA, typically as a consequence of safety or side effect issues."
    https://www.guidetopharmacology.org/helpPage.jsp

    RxNorm:
    - RXNORM_PRESCRIBABLE: "The RxNorm Current Prescribable Content is a subset of
    currently prescribable drugs found in RxNorm. We intend it to be an approximation of
    the prescription drugs currently marketed in the US. The subset also includes many
    over-the-counter drugs."
    https://www.nlm.nih.gov/research/umls/rxnorm/docs/prescribe.html
    """

    CHEMBL_NULL = "chembl_phase_null"
    CHEMBL_0_5 = "chembl_phase_0.5"
    CHEMBL_1 = "chembl_phase_1"
    CHEMBL_2 = "chembl_phase_2"
    CHEMBL_3 = "chembl_phase_3"
    CHEMBL_4 = "chembl_phase_4"
    CHEMBL_WITHDRAWN = "chembl_withdrawn"
    FDA_OTC = "fda_otc"
    FDA_PRESCRIPTION = "fda_prescription"
    FDA_DISCONTINUED = "fda_discontinued"
    FDA_TENTATIVE = "fda_tentative"
    HEMONC_APPROVED = "hemonc_approved"
    GTOPDB_APPROVED = "gtopdb_approved"
    GTOPDB_WITHDRAWN = "gtopdb_withdrawn"
    RXNORM_PRESCRIBABLE = "rxnorm_prescribable"


class HasIndication(BaseModel):
    """Data regarding disease indications from regulatory bodies."""

    disease_id: str
    disease_label: str
    normalized_disease_id: str | None = None
    supplemental_info: dict[str, str | None] | None = None

    model_config = ConfigDict(
        json_schema_extra={
            "example": {
                "disease_id": "mesh:D016778",
                "disease_label": "Malaria, Falciparum",
                "normalized_disease_id": "ncit:C34798",
                "supplemental_info": {"chembl_max_phase_for_ind": "chembl_phase_2"},
            }
        }
    )


class Therapy(BaseModel):
    """A pharmacologic substance used to treat a medical condition."""

    concept_id: str
    label: str | None = None
    aliases: list[str] | None = []
    trade_names: list[str] | None = []
    xrefs: list[str] | None = []
    associated_with: list[str] | None = []
    approval_ratings: list[ApprovalRating] | None = None
    approval_year: list[str] | None = []
    has_indication: list[HasIndication] | None = []

    model_config = ConfigDict(
        json_schema_extra={
            "example": {
                "label": "CISPLATIN",
                "concept_id": "chembl:CHEMBL11359",
                "aliases": [
                    "Cisplatin",
                    "Cis-Platinum II",
                    "Cisplatinum",
                    "cis-diamminedichloroplatinum(II)",
                    "CIS-DDP",
                    "INT-230-6 COMPONENT CISPLATIN",
                    "INT230-6 COMPONENT CISPLATIN",
                    "NSC-119875",
                    "Platinol",
                    "Platinol-Aq",
                ],
                "xrefs": [],
                "associated_with": None,
                "approval_ratings": "approved",
                "approval_year": [],
                "has_indication": [],
                "trade_names": ["PLATINOL", "PLATINOL-AQ", "CISPLATIN"],
            }
        }
    )


class RecordType(str, Enum):
    """Record item types."""

    IDENTITY = "identity"
    MERGER = "merger"


class RefType(str, Enum):
    """Reference item types."""

    # Must be in descending MatchType order.
    LABEL = "label"
    TRADE_NAMES = "trade_name"
    ALIASES = "alias"
    XREFS = "xref"
    ASSOCIATED_WITH = "associated_with"


# not incorporated as a RefType because it shouldn't be publicly searchable
RXNORM_BRAND_ITEM_TYPE = "rx_brand"


class MatchType(IntEnum):
    """Define string constraints for use in Match Type attributes."""

    CONCEPT_ID = 100
    LABEL = 80
    TRADE_NAME = 80
    ALIAS = 60
    XREF = 60
    ASSOCIATED_WITH = 60
    FUZZY_MATCH = 20
    NO_MATCH = 0


class SourcePriority(IntEnum):
    """Define constraints for Source Priority Rankings."""

    RXNORM = 1
    NCIT = 2
    HEMONC = 3
    DRUGBANK = 4
    DRUGSATFDA = 5
    GUIDETOPHARMACOLOGY = 6
    CHEMBL = 7
    CHEMIDPLUS = 8
    WIKIDATA = 9


class SourceName(str, Enum):
    """Define string constraints to ensure consistent capitalization."""

    WIKIDATA = "Wikidata"
    CHEMBL = "ChEMBL"
    NCIT = "NCIt"
    DRUGBANK = "DrugBank"
    CHEMIDPLUS = "ChemIDplus"
    RXNORM = "RxNorm"
    HEMONC = "HemOnc"
    DRUGSATFDA = "DrugsAtFDA"
    DRUGSATFDA_NDA = DRUGSATFDA
    DRUGSATFDA_ANDA = DRUGSATFDA  # noqa: PIE796
    GUIDETOPHARMACOLOGY = "GuideToPHARMACOLOGY"


class NamespacePrefix(Enum):
    """Define string constraints for namespace prefixes on concept IDs."""

    ATC = "atc"  # Anatomical Therapeutic Chemical Classification System
    CHEBI = "CHEBI"
    CHEMBL = "chembl"
    CHEMIDPLUS = "chemidplus"
    CASREGISTRY = CHEMIDPLUS
    CVX = "cvx"  # Vaccines Administered
    DRUGBANK = "drugbank"
    DRUGCENTRAL = "drugcentral"
    DRUGSATFDA_ANDA = "drugsatfda.anda"
    DRUGSATFDA_NDA = "drugsatfda.nda"
    HEMONC = "hemonc"
    INCHIKEY = "inchikey"
    IUPHAR_LIGAND = "iuphar.ligand"
    GUIDETOPHARMACOLOGY = IUPHAR_LIGAND
    MMSL = "mmsl"  # Multum MediSource Lexicon
    MSH = "mesh"  # Medical Subject Headings
    NCIT = "ncit"
    NDC = "ndc"  # National Drug Code
    PUBCHEMCOMPOUND = "pubchem.compound"
    PUBCHEMSUBSTANCE = "pubchem.substance"
    RXNORM = "rxcui"
    SPL = "spl"  # Structured Product Labeling
    UMLS = "umls"
    UNII = "unii"
    UNIPROT = "uniprot"
    USP = "usp"  # USP Compendial Nomenclature
    VANDF = "vandf"  # Veterans Health Administration National Drug File
    WIKIDATA = "wikidata"


# Source to URI. Will use OBO Foundry persistent URL (PURL) or source homepage
NAMESPACE_TO_SYSTEM_URI: dict[NamespacePrefix, str] = {
    NamespacePrefix.ATC: "https://www.who.int/tools/atc-ddd-toolkit/atc-classification/",
    NamespacePrefix.CHEBI: "http://purl.obolibrary.org/obo/chebi.owl",
    NamespacePrefix.CHEMBL: "https://www.ebi.ac.uk/chembl/",
    NamespacePrefix.CHEMIDPLUS: "https://pubchem.ncbi.nlm.nih.gov/source/ChemIDplus",
    NamespacePrefix.CASREGISTRY: "https://pubchem.ncbi.nlm.nih.gov/source/ChemIDplus",
    NamespacePrefix.CVX: "https://www2a.cdc.gov/vaccines/iis/iisstandards/vaccines.asp?rpt=cvx",
    NamespacePrefix.DRUGBANK: "https://go.drugbank.com",
    NamespacePrefix.DRUGCENTRAL: "https://drugcentral.org",
    NamespacePrefix.DRUGSATFDA_ANDA: "https://www.fda.gov/drugs/types-applications/abbreviated-new-drug-application-anda",
    NamespacePrefix.DRUGSATFDA_NDA: "https://www.fda.gov/drugs/types-applications/new-drug-application-nda",
    NamespacePrefix.HEMONC: "https://hemonc.org",
    NamespacePrefix.INCHIKEY: "https://www.chemspider.com",
    NamespacePrefix.IUPHAR_LIGAND: "https://www.guidetopharmacology.org/GRAC/LigandListForward",
    NamespacePrefix.GUIDETOPHARMACOLOGY: "https://www.guidetopharmacology.org/GRAC/LigandListForward",
    NamespacePrefix.MMSL: "https://www.nlm.nih.gov/research/umls/rxnorm/sourcereleasedocs/mmsl.html",
    NamespacePrefix.MSH: "https://id.nlm.nih.gov/mesh/",
    NamespacePrefix.NCIT: "http://purl.obolibrary.org/obo/ncit.owl",
    NamespacePrefix.NDC: "https://dps.fda.gov/ndc",
    NamespacePrefix.PUBCHEMCOMPOUND: "https://pubchem.ncbi.nlm.nih.gov/docs/compounds",
    NamespacePrefix.PUBCHEMSUBSTANCE: "https://pubchem.ncbi.nlm.nih.gov/docs/substances",
    NamespacePrefix.RXNORM: "https://www.nlm.nih.gov/research/umls/rxnorm/index.html",
    NamespacePrefix.SPL: "https://www.fda.gov/industry/fda-data-standards-advisory-board/structured-product-labeling-resources",
    NamespacePrefix.UMLS: "https://www.nlm.nih.gov/research/umls/index.html",
    NamespacePrefix.UNII: "https://precision.fda.gov/uniisearch",
    NamespacePrefix.UNIPROT: "https://www.uniprot.org",
    NamespacePrefix.USP: "https://www.usp.org/health-quality-safety/compendial-nomenclature",
    NamespacePrefix.VANDF: "https://www.nlm.nih.gov/research/umls/sourcereleasedocs/current/VANDF",
    NamespacePrefix.WIKIDATA: "https://www.wikidata.org",
}

# URI to source
SYSTEM_URI_TO_NAMESPACE = {
    system_uri: ns.value for ns, system_uri in NAMESPACE_TO_SYSTEM_URI.items()
}


class DataLicenseAttributes(BaseModel):
    """Define constraints for data license attributes."""

    non_commercial: StrictBool
    share_alike: StrictBool
    attribution: StrictBool


class ItemTypes(str, Enum):
    """Item types used in DynamoDB."""

    # Must be in descending MatchType order.
    LABEL = "label"
    TRADE_NAMES = "trade_name"
    ALIASES = "alias"
    XREFS = "xref"
    ASSOCIATED_WITH = "associated_with"


class SourceMeta(BaseModel):
    """Metadata for a given source to return in response object."""

    data_license: str
    data_license_url: str
    version: str
    data_url: str | None = None
    rdp_url: str | None = None
    data_license_attributes: dict[str, StrictBool]

    model_config = ConfigDict(
        json_schema_extra={
            "example": {
                "data_license": "CC BY-SA 3.0",
                "data_license_url": "https://creativecommons.org/licenses/by-sa/3.0/",
                "version": "27",
                "data_url": "http://ftp.ebi.ac.uk/pub/databases/chembl/ChEMBLdb/releases/chembl_27/",
                "rdp_url": "http://reusabledata.org/chembl.html",
                "data_license_attributes": {
                    "non_commercial": False,
                    "share_alike": True,
                    "attribution": True,
                },
            }
        }
    )


class SourceSearchMatches(BaseModel):
    """Container for matching information from an individual source.
    Used when matches are requested as an object, not an array.
    """

    match_type: MatchType
    records: list[Therapy]
    source_meta_: SourceMeta

    model_config = ConfigDict(
        json_schema_extra={
            "example": {
                "match_type": 80,
                "records": [
                    {
                        "concept_id": "chemidplus:15663-27-1",
                        "label": "Cisplatin",
                        "aliases": [
                            "1,2-Diaminocyclohexaneplatinum II citrate",
                            "cis-Diaminedichloroplatinum",
                        ],
                        "trade_names": [],
                        "xrefs": ["drugbank:DB00515"],
                        "associated_with": ["unii:Q20Q21Q62J"],
                        "approval_year": [],
                        "has_indication": [],
                    }
                ],
                "source_meta_": {
                    "data_license": "custom",
                    "data_license_url": "https://www.nlm.nih.gov/databases/download/terms_and_conditions.html",
                    "version": "20230222",
                    "data_url": "ftp://ftp.nlm.nih.gov/nlmdata/.chemidlease/",
                    "data_license_attributes": {
                        "non_commercial": False,
                        "attribution": True,
                        "share_alike": False,
                    },
                },
            }
        }
    )


class ServiceMeta(BaseModel):
    """Metadata regarding the therapy-normalization service."""

    name: Literal["thera-py"] = "thera-py"
    version: Literal[__version__] = __version__  # type: ignore[valid-type]
    response_datetime: datetime
    url: Literal["https://github.com/cancervariants/therapy-normalization"] = (
        "https://github.com/cancervariants/therapy-normalization"
    )

    model_config = ConfigDict(
        json_schema_extra={
            "example": {
                "name": "thera-py",
                "version": __version__,
                "response_datetime": "2021-04-05T16:44:15.367831",
                "url": "https://github.com/cancervariants/therapy-normalization",
            }
        }
    )


class MatchesNormalized(BaseModel):
    """Matches associated with normalized concept from a single source."""

    records: list[Therapy]
    source_meta_: SourceMeta


class BaseNormalizationService(BaseModel):
    """Base method providing shared attributes to Normalization service classes."""

    query: str
    warnings: list[dict] = []
    match_type: MatchType
    service_meta_: ServiceMeta


class UnmergedNormalizationService(BaseNormalizationService):
    """Response providing source records corresponding to normalization of user query.
    Enables retrieval of normalized concept while retaining sourcing for accompanying
    attributes.
    """

    normalized_concept_id: constr(pattern="^\\w[^:]*:.+$") | None = None  # type: ignore[valid-type]
    source_matches: dict[SourceName, MatchesNormalized]

    model_config = ConfigDict(
        json_schema_extra={
            "example": {
                "query": "L745870",
                "warnings": [],
                "match_type": 80,
                "service_meta_": {
                    "response_datetime": "2022-04-22T11:40:18.921859",
                    "name": "thera-py",
                    "version": __version__,
                    "url": "https://github.com/cancervariants/therapy-normalization",
                },
                "normalized_concept_id": "iuphar.ligand:3303",
                "source_matches": {
                    "GuideToPHARMACOLOGY": {
                        "records": [
                            {
                                "concept_id": "iuphar.ligand:3303",
                                "label": "L745870",
                                "aliases": [
                                    "L-745,870",
                                    "L 745870",
                                    "3-[[4-(4-chlorophenyl)piperazin-1-yl]methyl]-1H-pyrrolo[2,3-b]pyridine",
                                ],
                                "trade_names": [],
                                "xrefs": [
                                    "chemidplus:158985-00-3",
                                    "chembl:CHEMBL267014",
                                ],
                                "associated_with": [
                                    "pubchem.substance:178100340",
                                    "pubchem.compound:5311200",
                                    "inchikey:OGJGQVFWEPNYSB-UHFFFAOYSA-N",
                                ],
                                "approval_ratings": None,
                                "approval_year": [],
                                "has_indication": [],
                            }
                        ],
                        "source_meta_": {
                            "data_license": "CC BY-SA 4.0",
                            "data_license_url": "https://creativecommons.org/licenses/by-sa/4.0/",
                            "version": "2021.4",
                            "data_url": "https://www.guidetopharmacology.org/download.jsp",
                            "rdp_url": None,
                            "data_license_attributes": {
                                "non_commercial": False,
                                "share_alike": True,
                                "attribution": True,
                            },
                        },
                    },
                    "ChEMBL": {
                        "records": [
                            {
                                "concept_id": "chembl:CHEMBL267014",
                                "label": "L-745870",
                                "aliases": [],
                                "trade_names": [],
                                "xrefs": [],
                                "associated_with": [],
                                "approval_ratings": ["chembl_phase_0"],
                                "approval_year": [],
                                "has_indication": [],
                            }
                        ],
                        "source_meta_": {
                            "data_license": "CC BY-SA 3.0",
                            "data_license_url": "https://creativecommons.org/licenses/by-sa/3.0/",
                            "version": "29",
                            "data_url": "ftp://ftp.ebi.ac.uk/pub/databases/chembl/ChEMBLdb/releases/chembl_29",
                            "rdp_url": "http://reusabledata.org/chembl.html",
                            "data_license_attributes": {
                                "non_commercial": False,
                                "share_alike": True,
                                "attribution": True,
                            },
                        },
                    },
                },
            }
        }
    )


class NormalizationService(BaseNormalizationService):
    """Response containing one or more merged records and source data."""

    therapy: MappableConcept | None = None
    source_meta_: dict[SourceName, SourceMeta] | None = None

    model_config = ConfigDict(
        json_schema_extra={
            "example": {
                "query": "cisplatin",
                "warnings": None,
                "match_type": 80,
                "therapy": {
                    "conceptType": "Therapy",
                    "primaryCode": "rxcui:2555",
                    "id": "normalize.therapy.rxcui:2555",
                    "label": "cisplatin",
                    "mappings": [
                        {
                            "coding": {
                                "code": "2555",
                                "system": "https://www.nlm.nih.gov/research/umls/rxnorm/index.html",
                            },
                            "relation": "exactMatch",
                        },
                        {
                            "coding": {
                                "code": "C376",
                                "system": "http://purl.obolibrary.org/obo/ncit.owl",
                            },
                            "relation": "relatedMatch",
                        },
                        {
                            "coding": {
                                "code": "15663-27-1",
                                "system": "https://pubchem.ncbi.nlm.nih.gov/source/ChemIDplus",
                            },
                            "relation": "relatedMatch",
                        },
                        {
                            "coding": {"code": "Q412415", "system": "wikidata"},
                            "relation": "relatedMatch",
                        },
                        {
                            "coding": {"code": "L01XA01", "system": "atc"},
                            "relation": "relatedMatch",
                        },
                        {
                            "coding": {"code": "4456", "system": "mmsl"},
                            "relation": "relatedMatch",
                        },
                        {
                            "coding": {"code": "27899", "system": "chebi"},
                            "relation": "relatedMatch",
                        },
                        {
                            "coding": {"code": "5702198", "system": "pubchem.compound"},
                            "relation": "relatedMatch",
                        },
                        {
                            "coding": {"code": "C0008838", "system": "umls"},
                            "relation": "relatedMatch",
                        },
                        {
                            "coding": {"code": "m17910", "system": "usp"},
                            "relation": "relatedMatch",
                        },
                        {
                            "coding": {"code": "Q20Q21Q62J", "system": "fda"},
                            "relation": "relatedMatch",
                        },
                        {
                            "coding": {"code": "d00195", "system": "mmsl"},
                            "relation": "relatedMatch",
                        },
                        {
                            "coding": {"code": "Q20Q21Q62J", "system": "mthspl"},
                            "relation": "relatedMatch",
                        },
                        {
                            "coding": {"code": "31747", "system": "mmsl"},
                            "relation": "relatedMatch",
                        },
                        {
                            "coding": {"code": "D002945", "system": "mesh"},
                            "relation": "relatedMatch",
                        },
                        {
                            "coding": {"code": "4018139", "system": "vandf"},
                            "relation": "relatedMatch",
                        },
                    ],
                    "aliases": [
                        "CIS-DDP",
                        "cis Platinum",
                        "DDP",
                        "Dichlorodiammineplatinum",
                        "1,2-Diaminocyclohexaneplatinum II citrate",
                        "CISplatin",
                        "cis Diamminedichloroplatinum",
                        "CDDP",
                        "Diamminodichloride, Platinum",
                        "cis-Dichlorodiammineplatinum(II)",
                        "cis-Platinum",
                        "cis-diamminedichloroplatinum(II)",
                        "cis-Diamminedichloroplatinum(II)",
                        "Cis-DDP",
                        "cis-Diamminedichloroplatinum",
                        "cis-Diaminedichloroplatinum",
                        "Platinol-AQ",
                        "Platinol",
                        "Platinum Diamminodichloride",
                    ],
                    "extensions": [
                        {
                            "name": "trade_names",
                            "value": ["Platinol", "Cisplatin"],
                        },
                    ],
                },
                "source_meta_": {
                    "RxNorm": {
                        "data_license": "UMLS Metathesaurus",
                        "data_license_url": "https://www.nlm.nih.gov/research/umls/rxnorm/docs/termsofservice.html",
                        "version": "20210104",
                        "data_url": "https://www.nlm.nih.gov/research/umls/rxnorm/docs/rxnormfiles.html",
                        "rdp_url": None,
                        "data_license_attributes": {
                            "non_commercial": False,
                            "attribution": True,
                            "share_alike": False,
                        },
                    },
                    "NCIt": {
                        "data_license": "CC BY 4.0",
                        "data_license_url": "https://creativecommons.org/licenses/by/4.0/legalcode",
                        "version": "20.09d",
                        "data_url": "https://evs.nci.nih.gov/ftp1/NCI_Thesaurus/archive/20.09d_Release/",
                        "rdp_url": "http://reusabledata.org/ncit.html",
                        "data_license_attributes": {
                            "non_commercial": False,
                            "attribution": True,
                            "share_alike": False,
                        },
                    },
                    "ChemIDplus": {
                        "data_license": "custom",
                        "data_license_url": "https://www.nlm.nih.gov/databases/download/terms_and_conditions.html",
                        "version": "20200327",
                        "data_url": "ftp://ftp.nlm.nih.gov/nlmdata/.chemidlease/",
                        "rdp_url": None,
                        "data_license_attributes": {
                            "non_commercial": False,
                            "attribution": True,
                            "share_alike": False,
                        },
                    },
                    "Wikidata": {
                        "data_license": "CC0 1.0",
                        "data_license_url": "https://creativecommons.org/publicdomain/zero/1.0/",
                        "version": "20200812",
                        "data_url": None,
                        "rdp_url": None,
                        "data_license_attributes": {
                            "non_commercial": False,
                            "attribution": False,
                            "share_alike": False,
                        },
                    },
                },
                "service_meta_": {
                    "name": "thera-py",
                    "version": __version__,
                    "response_datetime": "2021-04-05T16:44:15.367831",
                    "url": "https://github.com/cancervariants/therapy-normalization",
                },
            }
        }
    )


class SearchService(BaseModel):
    """Core response schema containing matches for each source"""

    query: str
    warnings: list[dict] = []
    source_matches: dict[SourceName, SourceSearchMatches]
    service_meta_: ServiceMeta

    model_config = ConfigDict(
        json_schema_extra={
            "example": {
                "query": "cisplatin",
                "warnings": [],
                "source_matches": {
                    "ChemIDplus": {
                        "match_type": 80,
                        "records": [
                            {
                                "concept_id": "chemidplus:15663-27-1",
                                "label": "Cisplatin",
                                "aliases": [
                                    "1,2-Diaminocyclohexaneplatinum II citrate",
                                    "cis-Diaminedichloroplatinum",
                                ],
                                "trade_names": [],
                                "xrefs": ["drugbank:DB00515"],
                                "associated_with": ["unii:Q20Q21Q62J"],
                                "approval_year": [],
                                "has_indication": [],
                            }
                        ],
                        "source_meta_": {
                            "data_license": "custom",
                            "data_license_url": "https://www.nlm.nih.gov/databases/download/terms_and_conditions.html",
                            "version": "20230222",
                            "data_url": "ftp://ftp.nlm.nih.gov/nlmdata/.chemidlease/",
                            "data_license_attributes": {
                                "non_commercial": False,
                                "attribution": True,
                                "share_alike": False,
                            },
                        },
                    },
                    "GuideToPHARMACOLOGY": {
                        "match_type": 80,
                        "records": [
                            {
                                "concept_id": "iuphar.ligand:5343",
                                "label": "cisplatin",
                                "aliases": ["Platinol"],
                                "trade_names": [],
                                "xrefs": [
                                    "chembl:CHEMBL11359",
                                    "chemidplus:15663-27-1",
                                    "drugbank:DB00515",
                                ],
                                "associated_with": [
                                    "CHEBI:27899",
                                    "pubchem.compound:441203",
                                    "pubchem.substance:178102005",
                                ],
                                "approval_ratings": ["gtopdb_approved"],
                                "approval_year": [],
                                "has_indication": [],
                            }
                        ],
                        "source_meta_": {
                            "data_license": "CC BY-SA 4.0",
                            "data_license_url": "https://creativecommons.org/licenses/by-sa/4.0/",
                            "version": "2023.3",
                            "data_url": "https://www.guidetopharmacology.org/download.jsp",
                            "data_license_attributes": {
                                "non_commercial": False,
                                "attribution": True,
                                "share_alike": True,
                            },
                        },
                    },
                    "Wikidata": {
                        "match_type": 80,
                        "records": [
                            {
                                "concept_id": "wikidata:Q412415",
                                "label": "cisplatin",
                                "aliases": [
                                    "CDDP",
                                    "CIS-DDP",
                                    "Cis-DDP",
                                    "Platinol",
                                    "Platinol-AQ",
                                    "cis-diamminedichloroplatinum(II)",
                                ],
                                "trade_names": [],
                                "xrefs": [
                                    "chembl:CHEMBL11359",
                                    "chemidplus:15663-27-1",
                                    "drugbank:DB00515",
                                    "iuphar.ligand:5343",
                                    "rxcui:2555",
                                ],
                                "associated_with": ["pubchem.compound:5702198"],
                                "approval_year": [],
                                "has_indication": [],
                            }
                        ],
                        "source_meta_": {
                            "data_license": "CC0 1.0",
                            "data_license_url": "https://creativecommons.org/publicdomain/zero/1.0/",
                            "version": "20231212",
                            "data_license_attributes": {
                                "non_commercial": False,
                                "attribution": False,
                                "share_alike": False,
                            },
                        },
                    },
                    "DrugBank": {
                        "match_type": 80,
                        "records": [
                            {
                                "concept_id": "drugbank:DB00515",
                                "label": "Cisplatin",
                                "aliases": [
                                    "APRD00359",
                                    "CDDP",
                                    "Cis-DDP",
                                    "cis-diamminedichloroplatinum(II)",
                                    "cisplatino",
                                ],
                                "trade_names": [],
                                "xrefs": ["chemidplus:15663-27-1"],
                                "associated_with": [
                                    "inchikey:LXZZYRPGZAFOLE-UHFFFAOYSA-L",
                                    "unii:Q20Q21Q62J",
                                ],
                                "approval_year": [],
                                "has_indication": [],
                            }
                        ],
                        "source_meta_": {
                            "data_license": "CC0 1.0",
                            "data_license_url": "https://creativecommons.org/publicdomain/zero/1.0/",
                            "version": "5.1.10",
                            "data_url": "https://go.drugbank.com/releases/latest#open-data",
                            "rdp_url": "http://reusabledata.org/drugbank.html",
                            "data_license_attributes": {
                                "non_commercial": False,
                                "attribution": False,
                                "share_alike": False,
                            },
                        },
                    },
                    "HemOnc": {
                        "match_type": 80,
                        "records": [
                            {
                                "concept_id": "hemonc:105",
                                "label": "Cisplatin",
                                "aliases": [
                                    "CDDP",
                                    "DACP",
                                    "DDP",
                                    "NSC 119875",
                                    "NSC-119875",
                                    "NSC119875",
                                    "cis-diamminedichloroplatinum III",
                                    "cis-platinum",
                                    "cisplatinum",
                                ],
                                "trade_names": [],
                                "xrefs": ["rxcui:2555"],
                                "associated_with": [],
                                "approval_ratings": ["hemonc_approved"],
                                "approval_year": ["1978"],
                                "has_indication": [
                                    {
                                        "disease_id": "hemonc:569",
                                        "disease_label": "Bladder cancer",
                                        "normalized_disease_id": "ncit:C9334",
                                        "supplemental_info": {"regulatory_body": "FDA"},
                                    },
                                    {
                                        "disease_id": "hemonc:671",
                                        "disease_label": "Testicular cancer",
                                        "normalized_disease_id": "ncit:C7251",
                                        "supplemental_info": {"regulatory_body": "FDA"},
                                    },
                                ],
                            }
                        ],
                        "source_meta_": {
                            "data_license": "CC BY 4.0",
                            "data_license_url": "https://creativecommons.org/licenses/by/4.0/legalcode",
                            "version": "2023-09-05",
                            "data_url": "https://dataverse.harvard.edu/dataset.xhtml?persistentId=doi:10.7910/DVN/9CY9C6",
                            "data_license_attributes": {
                                "non_commercial": False,
                                "attribution": True,
                                "share_alike": False,
                            },
                        },
                    },
                    "NCIt": {
                        "match_type": 80,
                        "records": [
                            {
                                "concept_id": "ncit:C376",
                                "label": "Cisplatin",
                                "aliases": [],
                                "trade_names": [],
                                "xrefs": ["chemidplus:15663-27-1"],
                                "associated_with": [
                                    "CHEBI:27899",
                                    "umls:C0008838",
                                    "unii:Q20Q21Q62J",
                                ],
                                "approval_year": [],
                                "has_indication": [],
                            }
                        ],
                        "source_meta_": {
                            "data_license": "CC BY 4.0",
                            "data_license_url": "https://creativecommons.org/licenses/by/4.0/legalcode",
                            "version": "23.10e",
                            "data_url": "https://evs.nci.nih.gov/ftp1/NCI_Thesaurus/",
                            "rdp_url": "http://reusabledata.org/ncit.html",
                            "data_license_attributes": {
                                "non_commercial": False,
                                "attribution": True,
                                "share_alike": False,
                            },
                        },
                    },
                },
                "service_meta_": {
                    "name": "thera-py",
                    "version": __version__,
                    "response_datetime": "2021-04-05T16:44:15.367831",
                    "url": "https://github.com/cancervariants/therapy-normalization",
                },
            }
        }
    )



"""Main application for FastAPI"""

import html

from fastapi import FastAPI, HTTPException, Query
from fastapi.openapi.utils import get_openapi

from therapy import __version__
from therapy.database.database import create_db
from therapy.query import InvalidParameterError, QueryHandler
from therapy.schemas import (
    NormalizationService,
    SearchService,
    UnmergedNormalizationService,
)

query_handler = QueryHandler(create_db())
app = FastAPI(
    docs_url="/therapy",
    openapi_url="/therapy/openapi.json",
    swagger_ui_parameters={"tryItOutEnabled": True},
)


def custom_openapi() -> dict:
    """Generate custom fields for OpenAPI response"""
    if app.openapi_schema:
        return app.openapi_schema
    openapi_schema = get_openapi(
        title="The VICC Therapy Normalizer",
        version=__version__,
        openapi_version="3.0.3",
        description="Normalize drugs and other therapy terms.",
        routes=app.routes,
    )
    #    openapi_schema["info"]["license"] = {  # TODO
    #        "name": "Name-of-license",
    #        "url": "http://www.to-be-determined.com"
    #    }
    openapi_schema["info"]["contact"] = {
        "name": "Alex H. Wagner",
        "email": "Alex.Wagner@nationwidechildrens.org",
        "url": "https://www.nationwidechildrens.org/specialties/institute-for-genomic-medicine/research-labs/wagner-lab",
    }
    app.openapi_schema = openapi_schema
    return app.openapi_schema


app.openapi = custom_openapi  # type: ignore[method-assign]

# endpoint description text
get_matches_summary = "Given query, provide highest matches from each source."
response_descr = "A response to a validly-formed query."
q_descr = "Therapy to search."
incl_descr = (
    "Comma-separated list of source names to include in response. Will "
    "exclude all other sources. Will return HTTP status code 422: "
    'Unprocessable Entity if both "incl" and "excl" parameters are given.'
)
excl_descr = (
    "Comma-separated list of source names to exclude in response. Will "
    "include all other sources. Will return HTTP status code 422: "
    'Unprocessable Entity if both "incl" and "excl" parameters are given.'
)
infer_descr = (
    "If true, attempt namespace inference when queries match known "
    "Local Unique Identifier patterns."
)
search_description = (
    "For each source, return strongest-match concepts for query "
    "string provided by user"
)
normalize_description = (
    "Return merged strongest-match concept for query string " "provided by user."
)
merged_matches_summary = (
    "Given query, provide merged normalized record as a Therapy Mappable Concept."
)
merged_response_descr = "A response to a validly-formed query."
normalize_q_descr = "Therapy to normalize."
unmerged_matches_summary = (
    "Given query, provide source records corresponding to " "normalized concept."
)
unmerged_response_descr = (
    "Response containing source records contained within " "normalized concept."
)
unmerged_normalize_description = (
    "Return unmerged records associated with the "
    "normalized result of the user-provided query "
    "string."
)


@app.get(
    "/therapy/search",
    summary=get_matches_summary,
    operation_id="getQueryResponse",
    response_description=response_descr,
    response_model=SearchService,
    response_model_exclude_none=True,
    description=search_description,
)
def search(
    q: str = Query(..., description=q_descr),
    incl: str | None = Query(None, description=incl_descr),
    excl: str | None = Query(None, description=excl_descr),
    infer_namespace: bool = Query(True, description=infer_descr),
) -> SearchService:
    """For each source, return strongest-match concepts for query string
    provided by user.

    :param q: therapy search term
    :param str incl: comma-separated list of sources to include, with all others
        excluded. Raises HTTPException if both `incl` and `excl` are given.
    :param str excl: comma-separated list of sources exclude, with all others included.
        Raises HTTPException if both `incl` and `excl` are given.
    :param bool infer_namespace: if True, try to infer namespace from query term.
    :returns: JSON response with matched records and source metadata
    """
    try:
        response = query_handler.search(
            html.unescape(q),
            incl=incl,  # type: ignore[arg-type]
            excl=excl,  # type: ignore[arg-type]
            infer=infer_namespace,
        )
    except InvalidParameterError as e:
        raise HTTPException(status_code=422, detail=str(e)) from None
    return response


@app.get(
    "/therapy/normalize",
    summary=merged_matches_summary,
    operation_id="getMergedRecord",
    response_description=merged_response_descr,
    response_model=NormalizationService,
    response_model_exclude_none=True,
    description=normalize_description,
)
def normalize(
    q: str = Query(..., description=normalize_q_descr),
    infer_namespace: bool = Query(True, description=infer_descr),
) -> NormalizationService:
    """Return merged strongest-match concept for query string provided by
    user.

    :param q: therapy search term
    :param bool infer_namespace: if True, try to infer namespace from query term.
    :returns: JSON response with matching normalized record provided as a
    Therapy Mappable Concept, and source metadata
    """
    try:
        response = query_handler.normalize(html.unescape(q), infer_namespace)
    except InvalidParameterError as e:
        raise HTTPException(status_code=422, detail=str(e)) from None
    return response


@app.get(
    "/therapy/normalize_unmerged",
    summary=unmerged_matches_summary,
    operation_id="getUnmergedRecords",
    response_description=unmerged_response_descr,
    response_model=UnmergedNormalizationService,
    description=unmerged_normalize_description,
)
def normalize_unmerged(
    q: str = Query(..., description=normalize_q_descr),
    infer_namespace: bool = Query(True, description=infer_descr),
) -> UnmergedNormalizationService:
    """Return all individual records associated with a normalized concept.

    :param q: therapy search term
    :param bool infer_namespace: if True, try to infer namespace from query term.
    :returns: JSON response with matching normalized record and source metadata
    """
    try:
        response = query_handler.normalize_unmerged(html.unescape(q), infer_namespace)
    except InvalidParameterError as e:
        raise HTTPException(status_code=422, detail=str(e)) from None
    return response



"""Test ChemIDplus ETL methods."""

import isodate
import pytest

from therapy.etl.chemidplus import ChemIDplus
from therapy.schemas import MatchType, Therapy


@pytest.fixture(scope="module")
def chemidplus(test_source):
    """Provide test ChemIDplus query endpoint"""
    return test_source(ChemIDplus)


@pytest.fixture(scope="module")
def penicillin_v():
    """Build test fixture for Penicillin V drug."""
    return Therapy(
        concept_id="chemidplus:87-08-1",
        label="Penicillin V",
        aliases=["Phenoxymethylpenicillin"],
        trade_names=[],
        xrefs=[
            "drugbank:DB00417",
        ],
        associated_with=[
            "unii:Z61I075U2W",
        ],
    )


@pytest.fixture(scope="module")
def imatinib():
    """Build test fixture for Imatinib."""
    return Therapy(
        concept_id="chemidplus:152459-95-5",
        label="Imatinib",
        aliases=[],
        trade_names=[],
        xrefs=[
            "drugbank:DB00619",
        ],
        associated_with=[
            "unii:BKJ8M8G5HI",
        ],
    )


@pytest.fixture(scope="module")
def other_imatinib():
    """Build test fixture for imatinib mesylate."""
    return Therapy(
        concept_id="chemidplus:220127-57-1",
        label="Imatinib mesylate",
        aliases=[],
        xrefs=["drugbank:DB00619"],
        associated_with=["unii:8A1O1M485B"],
        trade_names=[],
    )


@pytest.fixture(scope="module")
def cisplatin():
    """Build test fixture for cisplatin."""
    return Therapy(
        concept_id="chemidplus:15663-27-1",
        label="Cisplatin",
        aliases=[
            "cis-Diaminedichloroplatinum",
            "1,2-Diaminocyclohexaneplatinum II citrate",
        ],
        trade_names=[],
        xrefs=["drugbank:DB00515"],
        associated_with=["unii:Q20Q21Q62J"],
    )


@pytest.fixture(scope="module")
def glycopyrronium_bromide():
    """Provide fixture for chemidplus:51186-83-5"""
    return Therapy(
        concept_id="chemidplus:51186-83-5",
        label="Glycopyrronium bromide",
        xrefs=["drugbank:DB00986"],
        associated_with=["unii:V92SO9WP2I"],
    )


def test_penicillin(chemidplus, compare_records, penicillin_v):
    """Test record retrieval for Penicillin V."""
    response = chemidplus.search("chemidplus:87-08-1")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], penicillin_v)

    response = chemidplus.search("CHemidplus:87-08-1")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], penicillin_v)

    response = chemidplus.search("Penicillin V")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 1
    compare_records(response.records[0], penicillin_v)

    response = chemidplus.search("Phenoxymethylpenicillin")
    assert response.match_type == MatchType.ALIAS
    assert len(response.records) == 1
    compare_records(response.records[0], penicillin_v)

    response = chemidplus.search("drugbank:DB00417")
    assert response.match_type == MatchType.XREF
    assert len(response.records) == 1
    compare_records(response.records[0], penicillin_v)

    response = chemidplus.search("unii:Z61I075U2W")
    assert response.match_type == MatchType.ASSOCIATED_WITH
    assert len(response.records) == 1
    compare_records(response.records[0], penicillin_v)

    response = chemidplus.search("87081")
    assert response.match_type == MatchType.NO_MATCH

    response = chemidplus.search("PenicillinV")
    assert response.match_type == MatchType.NO_MATCH


def test_imatinib(chemidplus, compare_records, imatinib, other_imatinib):
    """Test record retrieval for imatinib."""
    response = chemidplus.search("Imatinib")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 1
    compare_records(response.records[0], imatinib)

    response = chemidplus.search("imatiniB")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 1
    compare_records(response.records[0], imatinib)

    response = chemidplus.search("drugbank:DB00619")
    assert response.match_type == MatchType.XREF
    assert len(response.records) == 2
    for record in response.records:
        if record.concept_id == "chemidplus:152459-95-5":
            compare_records(record, imatinib)
        elif record.concept_id == "chemidplus:220127-57-1":
            compare_records(record, other_imatinib)
        else:
            pytest.fail("retrieved incorrect record")

    response = chemidplus.search("unii:BKJ8M8G5HI")
    assert response.match_type == MatchType.ASSOCIATED_WITH
    assert len(response.records) == 1
    compare_records(response.records[0], imatinib)


def test_cisplatin(chemidplus, compare_records, cisplatin):
    """Test record retrieval for cisplatin."""
    response = chemidplus.search("chemidplus:15663-27-1")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], cisplatin)

    response = chemidplus.search("Cisplatin")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 1
    compare_records(response.records[0], cisplatin)

    response = chemidplus.search("cis-Diaminedichloroplatinum")
    assert response.match_type == MatchType.ALIAS
    assert len(response.records) == 1
    compare_records(response.records[0], cisplatin)

    response = chemidplus.search("drugbank:DB00515")
    assert response.match_type == MatchType.XREF
    assert len(response.records) == 1
    compare_records(response.records[0], cisplatin)

    response = chemidplus.search("Cisplatine")
    assert response.match_type == MatchType.NO_MATCH


def test_glycopyrronium_bromide(chemidplus, compare_records, glycopyrronium_bromide):
    """Check input QC. This drug was processed with incorrect xref formatting."""
    response = chemidplus.search("chemidplus:51186-83-5")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], glycopyrronium_bromide)


def test_meta(chemidplus):
    """Test correctness of source metadata."""
    response = chemidplus.search("incoherent-string-of-text")
    assert response.source_meta_.data_license == "custom"
    assert (
        response.source_meta_.data_license_url
        == "https://www.nlm.nih.gov/databases/download/terms_and_conditions.html"
    )
    assert isodate.parse_date(response.source_meta_.version)
    assert (
        response.source_meta_.data_url == "ftp://ftp.nlm.nih.gov/nlmdata/.chemidlease/"
    )
    assert response.source_meta_.rdp_url is None
    assert response.source_meta_.data_license_attributes == {
        "non_commercial": False,
        "share_alike": False,
        "attribution": True,
    }



"""Construct test data for NCIt source."""

import xml.etree.ElementTree as XETree
from collections.abc import Generator
from pathlib import Path

import lxml.etree as etr
import owlready2 as owl
import xmlformatter

from therapy.database import create_db
from therapy.etl import NCIt

# define captured ids in `test_classes` variable

NCIT_PREFIX = "http://ncicb.nci.nih.gov/xml/owl/EVS/Thesaurus.owl"
OWL_PREFIX = "{http://www.w3.org/2002/07/owl#}"
RDFS_PREFIX = "{http://www.w3.org/2000/01/rdf-schema#}"
RDF_PREFIX = "{http://www.w3.org/1999/02/22-rdf-syntax-ns#}"

ANNOTATION_PROPERTY_TAG = f"{OWL_PREFIX}AnnotationProperty"
DESCRIPTION_TAG = f"{RDF_PREFIX}Description"
ABOUT_ATTRIB = f"{RDF_PREFIX}about"
AXIOM_TAG = f"{OWL_PREFIX}Axiom"
DATATYPE_TAG = f"{RDFS_PREFIX}Datatype"
OBJECT_PROPERTY_TAG = f"{OWL_PREFIX}ObjectProperty"
CLASS_TAG = f"{OWL_PREFIX}Class"


ncit = NCIt(create_db())
ncit._extract_data(False)
TEST_DATA_DIR = Path(__file__).resolve().parents[1] / "data" / "ncit"
outfile_path = TEST_DATA_DIR / ncit._data_file.name


def ncit_parser() -> Generator:
    """Get unique XML elements."""
    context = iter(
        etr.iterparse(ncit._data_file, events=("start", "end"), huge_tree=True)
    )
    for event, elem in context:
        if event == "start":
            yield elem


parser = ncit_parser()

# make root element
root = next(parser)
nsmap = root.nsmap
new_root = etr.Element(f"{RDF_PREFIX}RDF", nsmap=nsmap)
new_root.set(
    "{http://www.w3.org/XML/1998/namespace}base",
    "http://ncicb.nci.nih.gov/xml/owl/EVS/Thesaurus.owl",
)

# add ontology element
element = next(parser)
new_root.append(element)

# get annotation property elements
element = next(parser)
while element.tag != ANNOTATION_PROPERTY_TAG:
    element = next(parser)

while element.tag != DESCRIPTION_TAG:
    new_root.append(element)
    element = next(parser)
    while element.tag != ANNOTATION_PROPERTY_TAG and element.tag != DESCRIPTION_TAG:
        if element.tag == AXIOM_TAG:
            new_root.append(element)
        element = next(parser)

# get description elements
descriptions = {}
while element.tag != DATATYPE_TAG:
    new_root.append(element)
    element = next(parser)
    while element.tag != DESCRIPTION_TAG and element.tag != DATATYPE_TAG:
        if element.tag == AXIOM_TAG:
            new_root.append(element)
        element = next(parser)

# get datatype elements
while element.tag != OBJECT_PROPERTY_TAG:
    new_root.append(element)
    element = next(parser)
    while (
        element.tag != DATATYPE_TAG or len(element.attrib) == 0
    ) and element.tag != OBJECT_PROPERTY_TAG:
        if element.tag == AXIOM_TAG:
            new_root.append(element)
        element = next(parser)

# get object property elements
while element.tag != CLASS_TAG:
    new_root.append(element)
    element = next(parser)
    while element.tag != OBJECT_PROPERTY_TAG and element.tag != CLASS_TAG:
        if element.tag == AXIOM_TAG:
            new_root.append(element)
        element = next(parser)

# get classes
onto = owl.get_ontology(ncit._data_file.absolute().as_uri())
onto.load()
test_classes = {
    onto.C95221,
    onto.C74021,
    onto.C1647,
    onto.C49236,
    onto.C61796,
    onto.C376,
    onto.C739,
    onto.C839,
    onto.C107245,
    onto.C488,
    onto.C66724,
}

parent_concepts = set()
for c in test_classes:
    parent_concepts |= c.ancestors()  # type: ignore
parent_concepts.remove(owl.Thing)
parent_concept_iris = {p.iri for p in parent_concepts}

while element.attrib.get(ABOUT_ATTRIB) != f"{NCIT_PREFIX}#term-group":
    elements = [element]
    element = next(parser)
    while (
        element.tag != CLASS_TAG or ABOUT_ATTRIB not in element.attrib
    ) and element.attrib.get(ABOUT_ATTRIB) != f"{NCIT_PREFIX}#term-group":
        if element.tag == AXIOM_TAG:
            elements.append(element)
        element = next(parser)
    if elements[0].attrib.get(ABOUT_ATTRIB) in parent_concept_iris:
        for e in elements:
            new_root.append(e)
    else:
        for e in elements:
            e.clear()

# get annotations
EOF = False
while not EOF:
    new_root.append(element)
    element = next(parser)
    while True:
        if element.tag == AXIOM_TAG:
            new_root.append(element)
        try:
            element = next(parser)
        except StopIteration:
            EOF = True
            break
        if element.tag == DESCRIPTION_TAG:
            break

etr.ElementTree(new_root).write(outfile_path, pretty_print=True)

pi = XETree.ProcessingInstruction(  # TODO get encoding attrib out
    target='xml version="1.0"'
)
pi_string = XETree.tostring(pi).decode("ASCII")
with outfile_path.open("r+") as f:
    content = f.read()
    f.seek(0, 0)
    f.write(pi_string.rstrip("\r\n") + "\n" + content)

formatter = xmlformatter.Formatter(indent=2)
formatter.format_file(outfile_path)



"""ETL methods for ChemIDPlus® source.

Courtesy of the U.S. National Library of Medicine.
"""

import re
import xml.etree.ElementTree as ElTree
from collections.abc import Generator
from pathlib import Path

from rich.console import Console

from therapy.etl.base import Base
from therapy.schemas import (
    DataLicenseAttributes,
    NamespacePrefix,
    RecordParams,
    SourceMeta,
    SourceName,
)

TAGS_REGEX = r" \[.*\]"


class ChemIDplus(Base):
    """Class for ChemIDplus ETL methods."""

    @staticmethod
    def parse_xml(path: Path, tag: str) -> Generator:
        """Parse XML file and retrieve elements with matching tag value.
        :param Path path: path to XML file
        :param str tag: XML tag
        :return: generator yielding elements of corresponding tag
        """
        context = iter(ElTree.iterparse(path, events=("start", "end")))  # noqa: S314
        _, root = next(context)
        for event, elem in context:
            if event == "end" and elem.tag == tag:
                yield elem
                root.clear()

    def _transform_data(self) -> None:
        """Open dataset and prepare for loading into database."""
        parser = self.parse_xml(self._data_file, "Chemical")  # type: ignore
        # there's no elegant way to get a row total in advance from parsed XML --
        # use a rich Console to provide a static spinner instead
        with Console(color_system=None).status(
            "Loading ChemIDplus records...", spinner="dots"
        ):
            for chemical in parser:  # type: ignore
                if "displayName" not in chemical.attrib:
                    continue

                # initial setup and get label
                display_name = chemical.attrib["displayName"]
                if not display_name or not re.search(TAGS_REGEX, display_name):
                    continue
                label = re.sub(TAGS_REGEX, "", display_name)
                params: RecordParams = {"label": label}

                # get concept ID
                reg_no = chemical.find("NumberList").find("CASRegistryNumber")
                if not reg_no:
                    continue
                params["concept_id"] = (
                    f"{NamespacePrefix.CASREGISTRY.value}:{reg_no.text}"
                )

                # get aliases
                aliases = []
                label_l = label.lower()
                name_list = chemical.find("NameList")
                if name_list:
                    for name in name_list.findall("NameOfSubstance"):
                        text = name.text
                        if text != display_name and text.lower() != label_l:
                            aliases.append(re.sub(TAGS_REGEX, "", text))
                params["aliases"] = aliases

                # get xrefs and associated_with
                params["xrefs"] = []
                params["associated_with"] = []
                locator_list = chemical.find("LocatorList")
                if locator_list:
                    for loc in locator_list.findall("InternetLocator"):
                        if loc.text == "DrugBank":
                            db = (
                                f"{NamespacePrefix.DRUGBANK.value}:"
                                f"{loc.attrib['url'].split('/')[-1]}"
                            )
                            params["xrefs"].append(db)  # type: ignore
                        elif loc.text == "FDA SRS":
                            unii = (
                                f"{NamespacePrefix.UNII.value}:"
                                f"{loc.attrib['url'].split('/')[-1]}"
                            )
                            params["associated_with"].append(unii)  # type: ignore

                self._load_therapy(params)

    def _load_meta(self) -> None:
        """Add source metadata."""
        meta = SourceMeta(
            data_license="custom",
            data_license_url="https://www.nlm.nih.gov/databases/download/terms_and_conditions.html",
            version=self._version,
            data_url="ftp://ftp.nlm.nih.gov/nlmdata/.chemidlease/",
            rdp_url=None,
            data_license_attributes=DataLicenseAttributes(
                non_commercial=False, share_alike=False, attribution=True
            ).model_dump(),
        )
        self.database.add_source_metadata(SourceName.CHEMIDPLUS, meta)



"""Defines the RxNorm ETL methods.

"This product uses publicly available data courtesy of the U.S. National
Library of Medicine (NLM), National Institutes of Health, Department of Health
 and Human Services; NLM is not responsible for the product and does not
 endorse or recommend this or any other product."
"""

import csv
import logging
import re
from pathlib import Path

import yaml
from tqdm import tqdm
from wags_tails import CustomData, RxNormData

from therapy import ASSOC_WITH_SOURCES, ITEM_TYPES, XREF_SOURCES
from therapy.etl.base import Base
from therapy.schemas import (
    ApprovalRating,
    NamespacePrefix,
    RecordParams,
    SourceMeta,
    SourceName,
)

_logger = logging.getLogger(__name__)

# Designated Alias, Designated Syn, Tall Man Syn, Machine permutation
# Generic Drug Name, Designated Preferred Name, Preferred Entry Term,
# Clinical Drug, Entry Term, Rxnorm Preferred
ALIASES = ["SYN", "SY", "TMSY", "PM", "GN", "PT", "PEP", "CD", "ET", "RXN_PT"]

# Fully-specified drug brand name that can be prescribed
# Fully-specified drug brand name that can not be prescribed,
# Semantic branded drug
TRADE_NAMES = ["BD", "BN", "SBD"]

# Allowed rxnorm xrefs that have Source Level Restriction 0 or 1
RXNORM_XREFS = [
    "ATC",
    "CVX",
    "DRUGBANK",
    "MMSL",
    "MSH",
    "MTHCMSFRF",
    "MTHSPL",
    "RXNORM",
    "USP",
    "VANDF",
]


class RxNorm(Base):
    """Class for RxNorm ETL methods."""

    @staticmethod
    def _create_drug_form_yaml(drug_forms_file: Path, rxnorm_file: Path) -> None:
        """Create a YAML file containing RxNorm drug form values.

        :param drug_forms_file: location to save output file to
        :param rxnorm_file: source RRF file to extract drug form values from
        """
        dfs = []
        with rxnorm_file.open() as f:
            data = csv.reader(f, delimiter="|")
            for row in data:
                if ((row[12] == "DF") and (row[11] == "RXNORM")) and (
                    row[14] not in dfs
                ):
                    dfs.append(row[14])
        with drug_forms_file.open("w") as f:
            yaml.dump(dfs, f)

    _DataSourceClass = RxNormData

    def _extract_data(self, use_existing: bool) -> None:
        """Acquire source data.

        This method is responsible for initializing an instance of a data handler and,
        in most cases, setting ``self._data_file`` and ``self._version``.

        :param bool use_existing: if True, don't try to fetch latest source data
        """
        self._data_file, self._version = self._data_source.get_latest(
            from_local=use_existing
        )
        drug_form_data_handler = CustomData(
            src_name="rxnorm_drug_forms",
            filetype="yaml",
            latest_version_cb=lambda: self._version,
            download_cb=lambda version, file: self._create_drug_form_yaml(  # noqa: ARG005
                file,
                self._data_file,  # type: ignore
            ),
            data_dir=self._data_source.data_dir,
            file_name="rxnorm_drug_forms",
        )
        self._drug_forms_file, _ = drug_form_data_handler.get_latest()

    def _transform_data(self) -> None:
        """Transform the RxNorm source."""
        with self._drug_forms_file.open() as file:
            drug_forms = yaml.safe_load(file)

        with self._data_file.open() as f:  # type: ignore
            rff_data = csv.reader(f, delimiter="|")
            # Link ingredient to brand
            ingredient_brands: dict[str, str] = {}
            # Link precise ingredient to get brand
            precise_ingredient: dict[str, str] = {}
            # Transformed therapy records
            data: dict[str, dict] = {}
            # Link ingredient to brand
            sbdfs: dict[str, str] = {}
            # Get RXNORM|BN to concept_id
            brands: dict[str, str] = {}
            for row in rff_data:
                if row[11] in RXNORM_XREFS:
                    concept_id = f"{NamespacePrefix.RXNORM.value}:{row[0]}"
                    if row[12] == "BN" and row[11] == "RXNORM":
                        brands[row[14]] = concept_id
                    if row[12] == "SBDC" and row[11] == "RXNORM":
                        # Semantic Branded Drug Component
                        self._get_brands(row, ingredient_brands)
                    else:
                        if concept_id not in data:
                            params: RecordParams = {}
                            params["concept_id"] = concept_id
                            self._add_str_field(
                                params, row, precise_ingredient, drug_forms, sbdfs
                            )
                            self._add_xref_assoc(params, row)
                            data[concept_id] = params
                        else:
                            # Concept already created
                            params = data[concept_id]
                            self._add_str_field(
                                params, row, precise_ingredient, drug_forms, sbdfs
                            )
                            self._add_xref_assoc(params, row)

            for value in tqdm(data.values(), ncols=80, disable=self._silent):
                if "label" in value:
                    self._get_trade_names(
                        value, precise_ingredient, ingredient_brands, sbdfs
                    )
                    self._load_brand_concepts(value, brands)

                    params = {"concept_id": value["concept_id"]}

                    for field in [*list(ITEM_TYPES.keys()), "approval_ratings"]:
                        field_value = value.get(field)
                        if field_value:
                            params[field] = field_value
                    self._load_therapy(params)

    def _get_brands(self, row: list, ingredient_brands: dict) -> None:
        """Add ingredient and brand to ingredient_brands.

        :param List row: A row in the RxNorm data file.
        :param Dict ingredient_brands: Store brands for each ingredient
        """
        # SBDC: Ingredient(s) + Strength + [Brand Name]
        term = row[14]
        ingredients_brand = re.sub(
            r"(\d*)(\d*\.)?\d+ (MG|UNT|ML)?(/(ML|HR|MG))?", "", term
        )
        brand = term.split("[")[-1].split("]")[0]
        ingredients = ingredients_brand.replace(f"[{brand}]", "")
        if "/" in ingredients:
            ingredients_list = ingredients.split("/")
            for ingredient in ingredients_list:
                self._add_term(ingredient_brands, brand, ingredient.strip())
        else:
            self._add_term(ingredient_brands, brand, ingredients.strip())

    def _get_trade_names(
        self,
        value: dict,
        precise_ingredient: dict,
        ingredient_brands: dict,
        sbdfs: dict,
    ) -> None:
        """Get trade names for a given ingredient.

        :param Dict value: Therapy attributes
        :param Dict precise_ingredient: Brand names for precise ingredient
        :param Dict ingredient_brands: Brand names for ingredient
        :param Dict sbdfs: Brand names for ingredient from SBDF row
        """
        record_label = value["label"].lower()
        labels = [record_label]

        if "PIN" in value and value["PIN"] in precise_ingredient:
            for pin in precise_ingredient[value["PIN"]]:
                labels.append(pin.lower())  # noqa: PERF401

        for label in labels:
            trade_names: list[str] = [
                val for key, val in ingredient_brands.items() if label == key.lower()
            ]
            trade_names_uq = {val for sublist in trade_names for val in sublist}
            for tn in trade_names_uq:
                self._add_term(value, tn, "trade_names")

        if record_label in sbdfs:
            for tn in sbdfs[record_label]:
                self._add_term(value, tn, "trade_names")

    def _load_brand_concepts(self, rxnorm_record: dict, brands: dict) -> None:
        """Connect brand names to a concept and load into the database.

        :params rxnorm_record: A transformed therapy record
        :params brands: Connects brand names to concept records
        """
        if "trade_names" in rxnorm_record:
            for tn in rxnorm_record["trade_names"]:
                brand = brands.get(tn)
                if brand:
                    self.database.add_rxnorm_brand(brand, rxnorm_record["concept_id"])

    def _add_str_field(
        self,
        params: dict,
        row: list,
        precise_ingredient: dict,
        drug_forms: list,
        sbdfs: dict,
    ) -> None:
        """Differentiate STR field.

        :param Dict params: A transformed therapy record.
        :param List row: A row in the RxNorm data file.
        :param Dict precise_ingredient: Precise ingredient information
        :param List drug_forms: RxNorm Drug Form values
        :param Dict sbdfs: Brand names for precise ingredient
        """
        term = row[14]
        term_type = row[12]
        source = row[11]

        if (term_type == "IN" or term_type == "PIN") and source == "RXNORM":
            params["label"] = term
            if row[17] == "4096":
                params["approval_ratings"] = [ApprovalRating.RXNORM_PRESCRIBABLE.value]
        elif term_type in ALIASES:
            self._add_term(params, term, "aliases")
        elif term_type in TRADE_NAMES:
            self._add_term(params, term, "trade_names")

        if source == "RXNORM" and term_type == "SBDF":
            brand = term.split("[")[-1].split("]")[0]
            ingredient_strength = term.replace(f"[{brand}]", "")
            for df in drug_forms:
                if df in ingredient_strength:
                    ingredient = ingredient_strength.replace(df, "").strip()
                    self._add_term(sbdfs, brand, ingredient.lower())
                    break

        if source == "MSH":
            if term_type == "MH":
                # Get ID for accessing precise ingredient
                params["PIN"] = row[13]
            elif term_type == "PEP":
                self._add_term(precise_ingredient, term, row[13])

    @staticmethod
    def _add_term(params: dict, term: str, label_type: str) -> None:
        """Add a single term to a therapy record in an associated field.

        :param Dict params: A transformed therapy record.
        :param Str term: The term to be added
        :param Str label_type: The type of term
        """
        if params.get(label_type):
            if term not in params[label_type]:
                params[label_type].append(term)
        else:
            params[label_type] = [term]

    def _add_xref_assoc(self, params: dict, row: list) -> None:
        """Add xref or associated_with to therapy.

        :param Dict params: A transformed therapy record.
        :param List row: A row in the RxNorm data file.
        """
        ref = row[11]
        lui = row[13]
        if ref and lui != "NOCODE":
            xref_assoc = "UNII" if ref == "MTHSPL" else row[11].upper()

            if xref_assoc in XREF_SOURCES:
                source_id = f"{NamespacePrefix[xref_assoc].value}:{lui}"
                if source_id != params["concept_id"]:
                    # Sometimes concept_id is included in the source field
                    self._add_term(params, source_id, "xrefs")
            elif xref_assoc in ASSOC_WITH_SOURCES:
                source_id = f"{NamespacePrefix[xref_assoc].value}:{lui}"
                self._add_term(params, source_id, "associated_with")
            else:
                _logger.info("%s not in NameSpacePrefix.", xref_assoc)

    def _load_meta(self) -> None:
        """Add RxNorm metadata."""
        meta = SourceMeta(
            data_license="UMLS Metathesaurus",
            data_license_url="https://www.nlm.nih.gov/research/umls/rxnorm/docs/termsofservice.html",
            version=self._version,
            data_url="https://download.nlm.nih.gov/umls/kss/rxnorm/",
            rdp_url=None,
            data_license_attributes={
                "non_commercial": False,
                "share_alike": False,
                "attribution": True,
            },
        )
        self.database.add_source_metadata(SourceName.RXNORM, meta)



"""Module for testing DiseaseIndicationBase. This is needed so we know if
disease-normalizer introduced breaking changes.
"""

import os

import pytest
from disease.schemas import (
    SourceMeta as DiseaseSourceMeta,
)
from disease.schemas import (
    SourceName as DiseaseSourceName,
)

from therapy.database.database import AbstractDatabase
from therapy.etl import ChEMBL


def test_normalize_disease(is_test_env: bool, database: AbstractDatabase):
    """Test that DiseaseIndicationBase works correctly when normalizing diseases"""
    if not is_test_env:
        pytest.skip(
            "only perform direct check on Disease Normalizer in testing environment"
        )

    # set up normalizer
    os.environ["DISEASE_DYNAMO_TABLE"] = "disease_normalizer_therapy_test"
    chembl = ChEMBL(database)
    chembl.disease_normalizer.db.drop_db()
    chembl.disease_normalizer.db.initialize_db()
    chembl.disease_normalizer.db.add_source_metadata(
        DiseaseSourceName.MONDO,
        DiseaseSourceMeta(
            data_url="https://mondo.monarchinitiative.org/pages/download/",
            data_license="CC BY 4.0",
            data_license_url="https://creativecommons.org/licenses/by/4.0/legalcode",
            version="2022-10-11",
            rdp_url="http://reusabledata.org/monarch.html",
            data_license_attributes={
                "non_commercial": False,
                "attribution": True,
                "share_alike": False,
            },
        ),
    )
    chembl.disease_normalizer.db.add_record(
        {
            "concept_id": "mondo:0700110",
            "label": "pneumonia, non-human animal",
        },
        DiseaseSourceName.MONDO,
    )
    chembl.disease_normalizer.db.complete_write_transaction()
    result = chembl._normalize_disease("mondo:0700110")
    assert result == "mondo:0700110"



"""Build chemidplus test data."""

from collections.abc import Generator
from pathlib import Path
from xml.etree import ElementTree

from therapy.database import create_db
from therapy.etl import ChemIDplus

TEST_IDS = [
    "87-08-1",
    "152459-95-5",
    "220127-57-1",
    "15663-27-1",
    "50-06-6",
    "51186-83-5",
    "8025-81-8",
    "112901-68-5",
    "20537-88-6",
]

ch = ChemIDplus(create_db())
ch._extract_data(False)
TEST_DATA_DIR = Path(__file__).resolve().parents[1] / "data" / "chemidplus"
outfile_path = TEST_DATA_DIR / ch._data_file.name

root = ElementTree.Element("file")
root.set("name", ch._data_file.name)
root.set("date", ch._data_file.stem.split("chemidplus_")[1])


def parse_xml(path: Path) -> Generator:
    """Parse XML file and retrieve elements with matching tag value.
    :param Path path: path to XML file
    :param str tag: XML tag
    :return: generator yielding elements of corresponding tag
    """
    context = iter(ElementTree.iterparse(path, events=("start", "end")))  # noqa: S314
    _, root = next(context)
    for event, elem in context:
        if event == "end" and elem.tag == "Chemical":
            yield elem
            root.clear()


parser = parse_xml(ch._data_file)
for chemical in parser:
    regno = chemical.find("NumberList").find("CASRegistryNumber")
    if not regno:
        continue

    if regno.text in TEST_IDS:
        root.append(chemical)

with outfile_path.open("w") as f:
    ElementTree.ElementTree(root).write(f, encoding="unicode")

pi = ElementTree.ProcessingInstruction(
    target='xml version="1.0" encoding="UTF-8" standalone="yes"'
)
pi_string = ElementTree.tostring(pi).decode("UTF8")
with outfile_path.open("r+") as f:
    content = f.read()
    f.seek(0, 0)
    f.write(pi_string.rstrip("\r\n") + "\n" + content)
